# Copyright 2014, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Wrappers, constants, and helpers around ManagedSystem and its children."""
import re
import warnings

from oslo_log import log as logging

import pypowervm.const as c
from pypowervm.i18n import _
import pypowervm.util as u
import pypowervm.wrappers.entry_wrapper as ewrap
import pypowervm.wrappers.iocard as card
import pypowervm.wrappers.mtms as mtmwrap

LOG = logging.getLogger(__name__)

# ManagedSystem XPath constants
_PRIMARY_IP_ADDRESS = 'PrimaryIPAddress'
_HOST_IP_ADDRESS = _PRIMARY_IP_ADDRESS
_STATE = 'State'
_SYSTEM_NAME = 'SystemName'
_MASTER_MODE = 'IsPowerVMManagementMaster'

_SYS_CAPABILITIES = 'AssociatedSystemCapabilities'
_ACTIVE_LPM_CAP = u.xpath(
    _SYS_CAPABILITIES, 'ActiveLogicalPartitionMobilityCapable')
_INACTIVE_LPM_CAP = u.xpath(
    _SYS_CAPABILITIES, 'InactiveLogicalPartitionMobilityCapable')
_VETH_MAC_ADDR_CAP = u.xpath(
    _SYS_CAPABILITIES, 'VirtualEthernetCustomMACAddressCapable')
_IBMi_LPM_CAP = u.xpath(
    _SYS_CAPABILITIES, 'IBMiLogicalPartitionMobilityCapable')
_IBMi_RESTRICTEDIO_CAP = u.xpath(
    _SYS_CAPABILITIES, 'IBMiRestrictedIOModeCapable')
_SIMP_REMOTE_RESTART_CAP = u.xpath(
    _SYS_CAPABILITIES, 'PowerVMLogicalPartitionSimplifiedRemoteRestartCapable')
_AME_CAP = u.xpath(_SYS_CAPABILITIES, 'ActiveMemoryExpansionCapable')
_AIX_CAP = u.xpath(_SYS_CAPABILITIES, 'AIXCapable')
_IBMi_CAP = u.xpath(_SYS_CAPABILITIES, 'IBMiCapable')
_LINUX_CAP = u.xpath(_SYS_CAPABILITIES, 'LinuxCapable')
_SHR_PROC_POOL_CAP = u.xpath(
    _SYS_CAPABILITIES, 'SharedProcessorPoolCapable')
_VNIC_CAP = u.xpath(_SYS_CAPABILITIES, 'VirtualNICDedicatedSRIOVCapable')
_VNIC_FAILOVER_CAP = u.xpath(_SYS_CAPABILITIES, 'VirtualNICFailOverCapable')
_DYN_SRR_CAP = u.xpath(
    _SYS_CAPABILITIES, 'DynamicSimplifiedRemoteRestartToggleCapable')
_IBMi_NATIVE_IO_CAP = u.xpath(_SYS_CAPABILITIES, 'IBMiNativeIOCapable')
_DISABLE_SECURE_BOOT_CAP = u.xpath(
    _SYS_CAPABILITIES, 'DisableSecureBootCapable')

_SYS_MEM_CONFIG = 'AssociatedSystemMemoryConfiguration'
_MEMORY_INSTALLED = u.xpath(_SYS_MEM_CONFIG, 'InstalledSystemMemory')
_MEMORY_AVAIL = u.xpath(_SYS_MEM_CONFIG, 'CurrentAvailableSystemMemory')
_MEMORY_CONFIGURABLE = u.xpath(_SYS_MEM_CONFIG, 'ConfigurableSystemMemory')
_MEMORY_REGION_SIZE = u.xpath(_SYS_MEM_CONFIG, 'MemoryRegionSize')
_SYS_FIRMWARE_MEM = u.xpath(_SYS_MEM_CONFIG, 'MemoryUsedByHypervisor')
_PAGE_TABLE_RATIO = u.xpath(_SYS_MEM_CONFIG, 'DefaultHardwarePageTableRatio')

# Migration Constants
_SYS_PROC_CONFIG = 'AssociatedSystemProcessorConfiguration'
_PROC_COMPAT_MODES = u.xpath(
    _SYS_PROC_CONFIG, 'SupportedPartitionProcessorCompatibilityModes')
_MIN_PROC_UNITS_PER_CPU = u.xpath(
    _SYS_PROC_CONFIG, 'MinimumProcessorUnitsPerVirtualProcessor')
_MIGR_INFO = 'SystemMigrationInformation'
_MAX_ACTIVE_MIGR = u.xpath(_MIGR_INFO, 'MaximumActiveMigrations')
_MAX_INACTIVE_MIGR = u.xpath(_MIGR_INFO, 'MaximumInactiveMigrations')
_ACTIVE_MIGR_RUNNING = u.xpath(
    _MIGR_INFO, 'NumberOfActiveMigrationsInProgress')
_INACTIVE_MIGR_RUNNING = u.xpath(
    _MIGR_INFO, 'NumberOfInactiveMigrationsInProgress')
_MAX_FIRMWARE_MIGR = u.xpath(_MIGR_INFO, 'MaximumFirmwareActiveMigrations')

_PROC_UNITS_INSTALLED = u.xpath(
    _SYS_PROC_CONFIG, 'InstalledSystemProcessorUnits')

_PROC_UNITS_AVAIL = u.xpath(
    _SYS_PROC_CONFIG, 'CurrentAvailableSystemProcessorUnits')

_PROC_UNITS_CONFIGURABLE = u.xpath(
    _SYS_PROC_CONFIG, 'ConfigurableSystemProcessorUnits')

_MAX_PROCS_PER_PARTITION = u.xpath(
    _SYS_PROC_CONFIG, 'CurrentMaximumAllowedProcessorsPerPartition')

_MAX_PROCS_PER_AIX_LINUX_PARTITION = u.xpath(
    _SYS_PROC_CONFIG, 'CurrentMaximumProcessorsPerAIXOrLinuxPartition')

_MAX_VCPUS_PER_PARTITION = u.xpath(
    _SYS_PROC_CONFIG, 'MaximumAllowedVirtualProcessorsPerPartition')

_MAX_VCPUS_PER_AIX_LINUX_PARTITION = u.xpath(
    _SYS_PROC_CONFIG, 'CurrentMaximumVirtualProcessorsPerAIXOrLinuxPartition')

_VIOS_LINK = u.xpath("AssociatedVirtualIOServers", c.LINK)

# AssociatedSystemIOConfig constants
_ASIO_ROOT = 'AssociatedSystemIOConfiguration'
_ASIO_AVAIL_WWPNS = 'AvailableWWPNs'
_ASIO_HCA = 'HostChannelAdapters'
_ASIO_HEA = 'HostEthernetAdapters'
_ASIO_IOBUSES = 'IOBuses'
_ASIO_IOSLOTS = 'IOSlots'
_ASIO_SRIOVS = 'SRIOVAdapters'
_ASIO_ASVNET = 'AssociatedSystemVirtualNetwork'
_ASIO_WWPN_PREFIX = 'WWPNPrefix'

_IOSLOT_ROOT = 'IOSlot'
_IOSLOT_BUS_GRP_REQ = 'BusGroupingRequired'
_IOSLOT_DESC = 'Description'
_IOSLOT_FEAT_CODES = 'FeatureCodes'
_IOSLOT_PCI_CLASS = 'PCIClass'
_IOSLOT_PCI_DEV_ID = 'PCIDeviceID'
_IOSLOT_PCI_SUB_DEV_ID = 'PCISubsystemDeviceID'
_IOSLOT_PCI_REV_ID = 'PCIRevisionID'
_IOSLOT_PCI_VEND_ID = 'PCIVendorID'
_IOSLOT_PCI_SUB_VEND_ID = 'PCISubsystemVendorID'
_IOSLOT_DYN_REC_CON_INDEX = 'SlotDynamicReconfigurationConnectorIndex'
_IOSLOT_DYN_REC_CON_NAME = 'SlotDynamicReconfigurationConnectorName'


@ewrap.EntryWrapper.pvm_type('ManagedSystem')
class System(ewrap.EntryWrapper):
    """The PowerVM system that is being managed."""

    @property
    def system_name(self):
        return self._get_val_str(_SYSTEM_NAME)

    @property
    def mtms(self):
        return mtmwrap.MTMS.wrap(self.element.find(mtmwrap.MTMS_ROOT))

    @property
    def asio_config(self):
        return ASIOConfig.wrap(self.element.find(_ASIO_ROOT))

    @property
    def system_state(self):
        return self._get_val_str(_STATE, 'unknown')

    @property
    def proc_units(self):
        return self._get_val_float(_PROC_UNITS_INSTALLED, 0)

    @property
    def min_proc_units(self):
        return self._get_val_float(_MIN_PROC_UNITS_PER_CPU, 0)

    @property
    def proc_units_configurable(self):
        return self._get_val_float(_PROC_UNITS_CONFIGURABLE, 0)

    @property
    def proc_units_avail(self):
        return self._get_val_float(_PROC_UNITS_AVAIL, 0)

    @property
    def max_sys_procs_limit(self):
        return self._get_val_int(_MAX_PROCS_PER_PARTITION, 0)

    @property
    def max_procs_per_aix_linux_lpar(self):
        val = self._get_val_int(_MAX_PROCS_PER_AIX_LINUX_PARTITION, 0)
        # Some systems will not have maximum procs per lpar based on
        # partition type. In that case, use system max procs per partition.
        if val == 0:
            val = self.max_sys_procs_limit

        return val

    @max_procs_per_aix_linux_lpar.setter
    def max_procs_per_aix_linux_lpar(self, value):
        self.set_parm_value(_MAX_PROCS_PER_AIX_LINUX_PARTITION, str(value))

    @property
    def max_sys_vcpus_limit(self):
        return self._get_val_int(_MAX_VCPUS_PER_PARTITION, 0)

    @property
    def max_vcpus_per_aix_linux_lpar(self):
        val = self._get_val_int(_MAX_VCPUS_PER_AIX_LINUX_PARTITION, 0)
        # Some systems will not have maximum vcpus per lpar based on
        # partition type. In that case, use system max vcpus per partition.
        if val == 0:
            val = self.max_sys_vcpus_limit

        return val

    @max_vcpus_per_aix_linux_lpar.setter
    def max_vcpus_per_aix_linux_lpar(self, value):
        self.set_parm_value(_MAX_VCPUS_PER_AIX_LINUX_PARTITION, str(value))

    @property
    def memory_total(self):
        return self._get_val_int(_MEMORY_INSTALLED, 0)

    @property
    def memory_free(self):
        return self._get_val_int(_MEMORY_AVAIL, 0)

    @property
    def memory_configurable(self):
        return self._get_val_int(_MEMORY_CONFIGURABLE, 0)

    @property
    def memory_region_size(self):
        return self._get_val_int(_MEMORY_REGION_SIZE, 0)

    @property
    def firmware_memory(self):
        return self._get_val_int(_SYS_FIRMWARE_MEM, 0)

    @property
    def page_table_ratio(self):
        return self._get_val_int(_PAGE_TABLE_RATIO, 0)

    @property
    def host_ip_address(self):
        prop = _HOST_IP_ADDRESS
        val = self._get_val_str(prop)

        return val

    def _aix_capable(self):
        """Intepret AIX capability string and return True/False.

        Only expose this capability via externally via
        get_capabilities method.
        """
        cap = self._get_val_str(_AIX_CAP)
        # we can get 'unavailable' if PHYP interface is running an older
        # level and doesn't support query of this information
        if cap is not None and cap.lower() == 'inactive':
            return False
        # Default to true since that's what we always
        # assume prior to this capability
        return True

    def get_capabilities(self):
        """returns: The system capabilities from Power."""
        # VirtualEthernetCustomMACAddressCapable (custom_mac_addr_capable) will
        # default to True, which is the correct setting for POWER7 servers.
        # For operating system capabilities, they will be default to True
        # to match old Power server behaviors.
        cap_data = {'active_lpar_mobility_capable':
                    self._get_val_bool(_ACTIVE_LPM_CAP),
                    'active_memory_expansion_capable':
                    self._get_val_bool(_AME_CAP, False),
                    'inactive_lpar_mobility_capable':
                    self._get_val_bool(_INACTIVE_LPM_CAP),
                    'ibmi_lpar_mobility_capable':
                    self._get_val_bool(_IBMi_LPM_CAP, False),
                    'custom_mac_addr_capable':
                    self._get_val_bool(_VETH_MAC_ADDR_CAP, True),
                    'ibmi_restrictedio_capable':
                    self._get_val_bool(_IBMi_RESTRICTEDIO_CAP, False),
                    'ibmi_nativeio_capable':
                    self._get_val_bool(_IBMi_NATIVE_IO_CAP, False),
                    'simplified_remote_restart_capable':
                    self._get_val_bool(_SIMP_REMOTE_RESTART_CAP, False),
                    'aix_capable': self._aix_capable(),
                    'ibmi_capable':
                    self._get_val_bool(_IBMi_CAP, False),
                    'linux_capable':
                    self._get_val_bool(_LINUX_CAP, True),
                    'shared_processor_pool_capable':
                    self._get_val_bool(_SHR_PROC_POOL_CAP, False),
                    'dynamic_srr_capable':
                    self._get_val_bool(_DYN_SRR_CAP, False),
                    'vnic_capable': self._get_val_bool(_VNIC_CAP, False),
                    'vnic_failover_capable':
                    self._get_val_bool(_VNIC_FAILOVER_CAP, False),
                    'disable_secure_boot_capable':
                    self._get_val_bool(_DISABLE_SECURE_BOOT_CAP, False)}
        return cap_data

    @property
    def proc_compat_modes(self):
        """List of strings containing the processor compatibility modes.

        This is a READ-ONLY list.
        """
        return tuple(self._get_vals(_PROC_COMPAT_MODES))

    def highest_compat_mode(self):
        """This method returns the highest compatibility mode of the host."""
        modes = []
        pattern = r'^power(\d+)\+?$'
        for mode in self.proc_compat_modes:
            match = re.search(pattern, mode.lower())
            if match:
                modes.append(int(match.group(1)))
        modes = sorted(modes)
        if modes:
            return modes[-1]
        else:
            return 0

    @property
    def migration_data(self):
        """returns: The migration properties from PowerVM.

        This information should not be changed and should be treated as read
        only.
        """

        max_migr_sup = self._get_val_int(_MAX_FIRMWARE_MIGR)
        act_migr_sup = self._get_val_int(_MAX_ACTIVE_MIGR)
        inact_migr_sup = self._get_val_int(_MAX_INACTIVE_MIGR)
        pref_act_migr_sup = act_migr_sup
        pref_inact_migr_sup = inact_migr_sup
        act_migr_prog = self._get_val_int(_ACTIVE_MIGR_RUNNING)
        inact_migr_prog = self._get_val_int(_INACTIVE_MIGR_RUNNING)
        proc_compat = (','.join(self.proc_compat_modes))

        migr_data = {'max_migration_ops_supported': max_migr_sup,
                     'active_migrations_supported': act_migr_sup,
                     'inactive_migrations_supported': inact_migr_sup,
                     'preferred_active_migrations_supported':
                     pref_act_migr_sup,
                     'preferred_inactive_migrations_supported':
                     pref_inact_migr_sup,
                     'active_migrations_in_progress': act_migr_prog,
                     'inactive_migrations_in_progress': inact_migr_prog,
                     'proc_compat': proc_compat}

        # Copy get_capabilities() dictionary into migration_data in case
        # sometimes we need validate the host is capable for mobility.
        cap_data = self.get_capabilities()
        migr_data.update(cap_data)

        return migr_data

    @property
    def vios_links(self):
        """List of hrefs from AssociatedVirtualIOServers.

        This is a READ-ONLY list.
        """
        return self.get_href(_VIOS_LINK)

    @property
    def session_is_master(self):
        """The master mode state of this managed system.

        Use pypowervm.tasks.master_mode.request_master to request master mode
        :returns: True if the management node of this System's adapter.session
                  is the master.
        """
        return self._get_val_bool(_MASTER_MODE, True)


@ewrap.ElementWrapper.pvm_type(_ASIO_ROOT, has_metadata=True)
class ASIOConfig(ewrap.ElementWrapper):
    """The associated system IO configuration for this system."""

    @property
    def avail_wwpns(self):
        return self._get_val_int(_ASIO_AVAIL_WWPNS)

    @property
    def io_slots(self):
        es = ewrap.WrapperElemList(self._find_or_seed(_ASIO_IOSLOTS), IOSlot)
        return es

    @property
    def wwpn_prefix(self):
        return self._get_val_str(_ASIO_WWPN_PREFIX)

    @property
    def sriov_adapters(self):
        es = ewrap.WrapperElemList(self._find_or_seed(_ASIO_SRIOVS),
                                   child_class=card.SRIOVAdapter,
                                   indirect='IOAdapterChoice')
        return es


@ewrap.ElementWrapper.pvm_type(_IOSLOT_ROOT, has_metadata=True)
class IOSlot(ewrap.ElementWrapper):
    """An I/O Slot represents a device bus on the system.

    It may contain a piece of hardware within it.
    """

    @property
    def bus_grp_required(self):
        return self._get_val_bool(_IOSLOT_BUS_GRP_REQ)

    @property
    def description(self):
        return self._get_val_str(_IOSLOT_DESC)

    @property
    def feat_codes(self):
        return self._get_val_int(_IOSLOT_FEAT_CODES)

    @property
    def pci_class(self):
        return self._get_val_int(_IOSLOT_PCI_CLASS)

    @property
    def pci_dev_id(self):
        return self._get_val_int(_IOSLOT_PCI_DEV_ID)

    @property
    def pci_subsys_dev_id(self):
        return self._get_val_int(_IOSLOT_PCI_SUB_DEV_ID)

    @property
    def pci_sub_dev_id(self):
        """Deprecated - use pci_subsys_dev_id instead."""
        warnings.warn(_(
            "This property is deprecated!. "
            "Use pci_subsys_dev_id instead."), DeprecationWarning)
        return self.pci_subsys_dev_id

    @property
    def pci_rev_id(self):
        return self._get_val_int(_IOSLOT_PCI_REV_ID)

    @property
    def pci_revision_id(self):
        """Deprecated - use pci_rev_id instead."""
        warnings.warn(_(
            "This property is deprecated!. "
            "Use pci_rev_id instead."), DeprecationWarning)
        return self.pci_rev_id

    @property
    def pci_vendor_id(self):
        return self._get_val_int(_IOSLOT_PCI_VEND_ID)

    @property
    def pci_subsys_vendor_id(self):
        return self._get_val_int(_IOSLOT_PCI_SUB_VEND_ID)

    @property
    def pci_sub_vendor_id(self):
        """Deprecated - use pci_subsys_vendor_id instead."""
        warnings.warn(_(
            "This property is deprecated!. "
            "Use pci_subsys_vendor_id instead."), DeprecationWarning)
        return self.pci_subsys_vendor_id

    @property
    def drc_index(self):
        return self._get_val_int(_IOSLOT_DYN_REC_CON_INDEX)

    @property
    def dyn_reconfig_conn_index(self):
        """Deprecated - use drc_index instead."""
        warnings.warn(_(
            "This property is deprecated!. "
            "Use drc_index instead."), DeprecationWarning)
        return self.drc_index

    @property
    def drc_name(self):
        return self._get_val_str(_IOSLOT_DYN_REC_CON_NAME)

    @property
    def dyn_reconfig_conn_name(self):
        """Deprecated - use drc_name instead."""
        warnings.warn(_(
            "This property is deprecated!. "
            "Use drc_name instead."), DeprecationWarning)
        return self.drc_name



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock
import unittest

import pypowervm.adapter as adpt
import pypowervm.const as c
import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.storage as pvm_stor
import pypowervm.wrappers.virtual_io_server as vios


class TestVIOSWrapper(twrap.TestWrapper):

    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def test_update_timeout(self):
        self.adpt.update_by_path.return_value = self.dwrap.entry
        self.assertEqual(self.dwrap.entry, self.dwrap.update().entry)
        self.adpt.update_by_path.assert_called_with(self.dwrap, None, mock.ANY,
                                                    timeout=3600)
        self.assertEqual(self.dwrap.entry, self.dwrap.update(timeout=42).entry)
        self.adpt.update_by_path.assert_called_with(self.dwrap, None, mock.ANY,
                                                    timeout=42)
        # If the session is configured for longer...
        self.adpt.session.timeout = 10000
        self.assertEqual(self.dwrap.entry, self.dwrap.update().entry)
        # ...default to the longer value.
        self.adpt.update_by_path.assert_called_with(self.dwrap, None, mock.ANY,
                                                    timeout=10000)
        # But explicit timeout can still be set.
        self.assertEqual(self.dwrap.entry, self.dwrap.update(timeout=42).entry)
        self.adpt.update_by_path.assert_called_with(self.dwrap, None, mock.ANY,
                                                    timeout=42)

    def test_get_ip_addresses(self):
        expected_ips = ('9.1.2.4', '10.10.10.5')
        self.assertEqual(expected_ips, self.dwrap.ip_addresses)

    def test_mover_service_partition(self):
        self.assertTrue(self.dwrap.is_mover_service_partition)
        self.dwrap.is_mover_service_partition = False
        self.assertFalse(self.dwrap.is_mover_service_partition)

    def test_rmc_ip(self):
        self.assertEqual('9.1.2.5', self.dwrap.rmc_ip)

    def test_license_accept(self):
        self.assertTrue(self.dwrap.is_license_accepted)

    def test_vnic_capabilities(self):
        self.assertTrue(self.dwrap.vnic_capable)
        self.assertTrue(self.dwrap.vnic_failover_capable)

    def test_hdisk_reserve_policy_found(self):
        # Most are NoReserve; look for the only one that's SinglePath to make
        # sure we're actually searching rather than picking first/last/random
        found_policy = self.dwrap.hdisk_reserve_policy(
            '6005076300838041300000000000002B')
        self.assertEqual('SinglePath', found_policy)

    def test_hdisk_reserve_policy_notfound(self):
        # Most are NoReserve; look for the only one that's SinglePath to make
        # sure we're actually searching rather than picking first/last/random
        found_policy = self.dwrap.hdisk_reserve_policy('Bogus')
        self.assertIsNone(found_policy)

    def test_hdisk_from_uuid_found(self):
        found_name = self.dwrap.hdisk_from_uuid(
            '01M0lCTTIxNDUyNEM2MDA1MDc2MzAwODM4MDQxMzAwMDAwMDAwMDAwMDhCNQ==')
        self.assertEqual('hdisk7', found_name)

    def test_hdisk_from_uuid_notfound(self):
        found_name = self.dwrap.hdisk_from_uuid('Bogus')
        self.assertIsNone(found_name)

    def test_seas(self):
        self.assertEqual(1, len(self.dwrap.seas))
        sea = self.dwrap.seas[0]
        self.assertEqual(1, sea.pvid)
        self.assertEqual(1, len(sea.addl_adpts))

    def test_trunks(self):
        self.assertEqual(3, len(self.dwrap.trunk_adapters))
        self.assertEqual(1, self.dwrap.trunk_adapters[0].pvid)
        self.assertEqual(4094, self.dwrap.trunk_adapters[1].pvid)
        self.assertEqual(4093, self.dwrap.trunk_adapters[2].pvid)

    def test_derive_orphan_trunk_adapters(self):
        orphans = self.dwrap.derive_orphan_trunk_adapters()
        self.assertEqual(1, len(orphans))
        self.assertEqual(4093, orphans[0].pvid)

    def test_wwpns(self):
        """Tests the helper methods to get WWPNs more easily."""
        phys_paths = self.dwrap.get_pfc_wwpns()
        self.assertIsNotNone(phys_paths)
        self.assertEqual(2, len(phys_paths))

        virt_paths = self.dwrap.get_vfc_wwpns()
        self.assertIsNotNone(virt_paths)
        self.assertEqual(2, len(virt_paths))
        for virt_path in virt_paths:
            self.assertEqual(2, len(virt_path))

        self.assertEqual(1, len(self.dwrap.get_active_pfc_wwpns()))
        self.assertEqual('10000090FA1B6302',
                         self.dwrap.get_active_pfc_wwpns()[0])

    def test_pfc_ports(self):
        """Tests that the physical FC ports can be gathered."""
        ports = self.dwrap.pfc_ports
        self.assertIsNotNone(ports)
        self.assertEqual(2, len(ports))

        # Validate attributes on one.
        self.assertEqual('U78AB.001.WZSJBM3-P1-C2-T2', ports[0].loc_code)
        self.assertEqual('fcs1', ports[0].name)
        self.assertEqual('1aU78AB.001.WZSJBM3-P1-C2-T2', ports[0].udid)
        self.assertEqual('10000090FA1B6303', ports[0].wwpn)
        self.assertEqual(0, ports[0].npiv_available_ports)
        self.assertEqual(0, ports[0].npiv_total_ports)

    def test_phys_vols(self):
        """Tests that the physical volumes can be gathered."""
        phys_vols = self.dwrap.phys_vols
        self.assertIsNotNone(phys_vols)
        self.assertEqual(11, len(phys_vols))

        # Validate attributes on one.
        self.assertEqual(phys_vols[0].description, 'SAS Disk Drive')
        self.assertEqual(phys_vols[0].udid,
                         '01M0lCTU1CRjI2MDBSQzUwMDAwMzk0NzgzQTUyQjg=')
        self.assertEqual(phys_vols[0].capacity, 572325)
        self.assertEqual(phys_vols[0].name, 'hdisk0')
        self.assertEqual(phys_vols[0].state, 'active')


class TestViosMappings(twrap.TestWrapper):

    file = 'fake_vios_mappings.txt'
    wrapper_class_to_test = vios.VIOS
    mock_adapter_fx_args = {}

    def setUp(self):
        super(TestViosMappings, self).setUp()
        self.adpt.build_href.return_value = "a_link"

    def test_bld_scsi_mapping_vopt(self):
        """Validation that the element is correct."""
        vopt = pvm_stor.VOptMedia.bld_ref(self.adpt, 'media_name')
        vmap = vios.VSCSIMapping.bld(self.adpt, 'host_uuid',
                                     'client_lpar_uuid', vopt)
        self.assertIsNotNone(vmap)
        self.assertIsNotNone(vmap.element)
        self.assertEqual(vmap.client_adapter.side, 'Client')
        self.assertTrue(vmap.client_adapter._get_val_bool(
            'UseNextAvailableSlotID'))
        self.assertEqual(vmap.server_adapter.side, 'Server')
        # Validate the exact XML of the server adapter: ensure proper ordering.
        self.assertEqual(
            '<uom:ServerAdapter xmlns:uom="http://www.ibm.com/xmlns/systems/po'
            'wer/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata>'
            '<uom:Atom/></uom:Metadata><uom:AdapterType>Server</uom:AdapterTyp'
            'e><uom:UseNextAvailableSlotID>true</uom:UseNextAvailableSlotID></'
            'uom:ServerAdapter>'.encode('utf-8'),
            vmap.server_adapter.toxmlstring())
        # If the slot number is None then REST will assign the first available.
        self.assertIsNone(vmap.client_adapter.lpar_slot_num)
        self.assertIsNone(vmap.target_dev)
        self.assertEqual('media_name', vmap.backing_storage.media_name)
        self.assertEqual('a_link', vmap.client_lpar_href)
        self.assertIsInstance(vmap.backing_storage, pvm_stor.VOptMedia)

        # Test the cloning
        vopt2 = pvm_stor.VOptMedia.bld_ref(self.adpt, 'media_name2')
        vmap2 = vios.VSCSIMapping.bld_from_existing(vmap, vopt2)
        self.assertIsNotNone(vmap2)
        self.assertIsNotNone(vmap2.element)
        self.assertEqual(vmap2.client_adapter.side, 'Client')
        self.assertEqual(vmap2.server_adapter.side, 'Server')
        self.assertIsNone(vmap2.client_adapter.lpar_slot_num)
        self.assertIsNone(vmap2.target_dev)
        self.assertEqual('media_name2', vmap2.backing_storage.media_name)
        self.assertEqual('a_link', vmap2.client_lpar_href)
        self.assertIsInstance(vmap2.backing_storage, pvm_stor.VOptMedia)

        # Clone to a different device type
        vdisk = pvm_stor.VDisk.bld_ref(self.adpt, 'disk_name')
        vmap3 = vios.VSCSIMapping.bld_from_existing(
            vmap, vdisk, lpar_slot_num=6, lua='vdisk_lua')
        self.assertIsNotNone(vmap3)
        self.assertIsNotNone(vmap3.element)
        # Validate the exact XML of the client adapter: ensure proper ordering.
        self.assertEqual(
            '<uom:ClientAdapter xmlns:uom="http://www.ibm.com/xmlns/systems/po'
            'wer/firmware/uom/mc/2012_10/" schemaVersion="V1_0"><uom:Metadata>'
            '<uom:Atom/></uom:Metadata><uom:AdapterType>Client</uom:AdapterTyp'
            'e><uom:UseNextAvailableSlotID>false</uom:UseNextAvailableSlotID><'
            'uom:VirtualSlotNumber>6</uom:VirtualSlotNumber></uom:ClientAdapte'
            'r>'.encode('utf-8'), vmap3.client_adapter.toxmlstring())
        self.assertEqual('Client', vmap3.client_adapter.side)
        # Specifying 'lua' builds the appropriate type of target dev...
        self.assertIsInstance(vmap3.target_dev, pvm_stor.VDiskTargetDev)
        # ...with the correct LUA
        self.assertEqual('vdisk_lua', vmap3.target_dev.lua)
        self.assertEqual(6, vmap3.client_adapter.lpar_slot_num)
        # Assert this is set to False when specifying the slot number
        # and building from an existing mapping
        self.assertFalse(vmap3.client_adapter._get_val_bool(
            'UseNextAvailableSlotID'))
        self.assertEqual('Server', vmap3.server_adapter.side)
        self.assertEqual('disk_name', vmap3.backing_storage.name)
        self.assertEqual('a_link', vmap3.client_lpar_href)
        self.assertIsInstance(vmap3.backing_storage, pvm_stor.VDisk)

    def test_bld_scsi_mapping_vdisk(self):
        """Validation that the element is correct."""
        vdisk = pvm_stor.VDisk.bld_ref(self.adpt, 'disk_name')
        vmap = vios.VSCSIMapping.bld(self.adpt, 'host_uuid',
                                     'client_lpar_uuid', vdisk,
                                     lpar_slot_num=5, lua='vdisk_lua')
        self.assertIsNotNone(vmap)
        self.assertIsNotNone(vmap.element)
        self.assertEqual('Client', vmap.client_adapter.side)
        self.assertIsInstance(vmap.target_dev, pvm_stor.VDiskTargetDev)
        self.assertEqual('vdisk_lua', vmap.target_dev.lua)
        self.assertEqual(5, vmap.client_adapter.lpar_slot_num)
        # Assert that we set this to False when specifying the slot number
        self.assertFalse(vmap.client_adapter._get_val_bool(
            'UseNextAvailableSlotID'))
        self.assertEqual('Server', vmap.server_adapter.side)
        self.assertEqual('disk_name', vmap.backing_storage.name)
        self.assertEqual('a_link', vmap.client_lpar_href)
        self.assertIsInstance(vmap.backing_storage, pvm_stor.VDisk)

        # Test cloning
        vdisk2 = pvm_stor.VDisk.bld_ref(self.adpt, 'disk_name2')
        vmap2 = vios.VSCSIMapping.bld_from_existing(vmap, vdisk2,
                                                    lpar_slot_num=6)
        self.assertIsNotNone(vmap2)
        self.assertIsNotNone(vmap2.element)
        self.assertEqual('Client', vmap2.client_adapter.side)
        # Cloning without specifying 'lua' doesn't clone the target dev
        self.assertIsNone(vmap2.target_dev)
        self.assertEqual(6, vmap2.client_adapter.lpar_slot_num)
        self.assertFalse(vmap2.client_adapter._get_val_bool(
            'UseNextAvailableSlotID'))
        self.assertEqual('Server', vmap2.server_adapter.side)
        self.assertEqual('disk_name2', vmap2.backing_storage.name)
        self.assertEqual('a_link', vmap2.client_lpar_href)
        self.assertIsInstance(vmap2.backing_storage, pvm_stor.VDisk)

    def test_bld_scsi_mapping_lu(self):
        """Validation that the element is correct."""
        lu = pvm_stor.LU.bld_ref(self.adpt, 'disk_name', 'udid')
        vmap = vios.VSCSIMapping.bld(self.adpt, 'host_uuid',
                                     'client_lpar_uuid', lu,
                                     lpar_slot_num=5)
        self.assertIsNotNone(vmap)
        self.assertIsNotNone(vmap.element)
        self.assertEqual('Client', vmap.client_adapter.side)
        self.assertIsNone(vmap.target_dev)
        self.assertEqual(5, vmap.client_adapter.lpar_slot_num)
        self.assertEqual('Server', vmap.server_adapter.side)
        self.assertEqual('disk_name', vmap.backing_storage.name)
        self.assertEqual('udid', vmap.backing_storage.udid)
        self.assertEqual('a_link', vmap.client_lpar_href)
        self.assertIsInstance(vmap.backing_storage, pvm_stor.LU)

        # Test cloning
        lu2 = pvm_stor.LU.bld_ref(self.adpt, 'disk_name2', 'udid2')
        vmap2 = vios.VSCSIMapping.bld_from_existing(vmap, lu2, lua='lu_lua')
        self.assertIsNotNone(vmap2)
        self.assertIsNotNone(vmap2.element)
        self.assertEqual('Client', vmap2.client_adapter.side)
        self.assertEqual(5, vmap2.client_adapter.lpar_slot_num)
        self.assertIsInstance(vmap2.target_dev, pvm_stor.LUTargetDev)
        self.assertEqual('lu_lua', vmap2.target_dev.lua)
        self.assertEqual('Server', vmap2.server_adapter.side)
        self.assertEqual('disk_name2', vmap2.backing_storage.name)
        self.assertEqual('udid2', vmap2.backing_storage.udid)
        self.assertEqual('a_link', vmap2.client_lpar_href)
        self.assertIsInstance(vmap2.backing_storage, pvm_stor.LU)

    def test_bld_scsi_mapping_pv(self):
        """Validation that the element is correct."""
        pv = pvm_stor.PV.bld(self.adpt, 'disk_name', 'udid')
        vmap = vios.VSCSIMapping.bld(self.adpt, 'host_uuid',
                                     'client_lpar_uuid', pv,
                                     lpar_slot_num=5, target_name='fake_name')
        self.assertIsNotNone(vmap)
        self.assertIsNotNone(vmap.element)
        self.assertEqual('Client', vmap.client_adapter.side)
        self.assertEqual(5, vmap.client_adapter.lpar_slot_num)
        self.assertEqual('Server', vmap.server_adapter.side)
        self.assertEqual('disk_name', vmap.backing_storage.name)
        self.assertEqual('a_link', vmap.client_lpar_href)
        self.assertEqual('fake_name', vmap.target_dev.name)
        self.assertIsInstance(vmap.backing_storage, pvm_stor.PV)

        # Test cloning
        pv2 = pvm_stor.PV.bld(self.adpt, 'disk_name2', 'udid2')
        vmap2 = vios.VSCSIMapping.bld_from_existing(
            vmap, pv2, lpar_slot_num=6, lua='pv_lua')
        self.assertIsNotNone(vmap2)
        self.assertIsNotNone(vmap2.element)
        self.assertEqual('Client', vmap2.client_adapter.side)
        self.assertEqual(6, vmap2.client_adapter.lpar_slot_num)
        self.assertIsInstance(vmap2.target_dev, pvm_stor.PVTargetDev)
        self.assertEqual('pv_lua', vmap2.target_dev.lua)
        self.assertEqual('Server', vmap2.server_adapter.side)
        self.assertEqual('disk_name2', vmap2.backing_storage.name)
        self.assertEqual('a_link', vmap2.client_lpar_href)
        self.assertIsNone(vmap2.target_dev.name)
        self.assertIsInstance(vmap2.backing_storage, pvm_stor.PV)

        # Test empty target_dev_type
        pv3 = pvm_stor.PV.bld(self.adpt, 'disk_name3', 'udid3')
        vmap3 = vios.VSCSIMapping.bld_from_existing(
            vmap, pv3, lpar_slot_num=6)
        self.assertIsNone(vmap3.target_dev)

    def test_clone_scsi_mapping_no_storage(self):
        """Clone a VSCSI mapping with no storage element."""
        pv = pvm_stor.PV.bld(self.adpt, 'disk_name', 'udid')
        vmap = vios.VSCSIMapping.bld(self.adpt, 'host_uuid',
                                     'client_lpar_uuid', pv,
                                     lpar_slot_num=5)
        vmap2 = vios.VSCSIMapping.bld_from_existing(vmap, None)
        self.assertIsNotNone(vmap2)
        self.assertIsNotNone(vmap2.element)
        self.assertEqual('Client', vmap2.client_adapter.side)
        self.assertEqual('Server', vmap2.server_adapter.side)
        self.assertEqual('a_link', vmap2.client_lpar_href)
        self.assertEqual(5, vmap2.client_adapter.lpar_slot_num)
        self.assertIsNone(vmap.target_dev)
        self.assertIsNone(vmap2.backing_storage)
        # Illegal to specify target dev properties without backing storage.
        self.assertRaises(ValueError, vios.VSCSIMapping.bld_from_existing,
                          vmap, None, lua='bogus')

    def test_get_scsi_mappings(self):
        mappings = self.dwrap.scsi_mappings

        # Ensure that at least one adapter has a client LPAR & storage
        found_client_uri = False
        static_map = None
        for mapping in mappings:
            if mapping.client_lpar_href and mapping.backing_storage:
                found_client_uri = True
                static_map = mapping
        self.assertTrue(found_client_uri)

        # We'll use the previous mapping as a baseline for further validation
        self.assertIsNotNone(static_map.client_adapter)
        self.assertIsNotNone(static_map.backing_storage)
        self.assertIsNotNone(static_map.server_adapter)

        # Deeper check on each of these.
        ca = static_map.client_adapter
        self.assertEqual(5, ca.lpar_id)
        self.assertEqual(1, ca.vios_id)
        self.assertTrue(ca.is_varied_on)
        self.assertIsNotNone(ca.lpar_slot_num)
        self.assertIsNotNone(ca.vios_slot_num)
        self.assertIsNotNone(ca.loc_code)
        self.assertEqual(ca.side, 'Client')

        sa = static_map.server_adapter
        self.assertEqual(10, sa.lpar_id)
        self.assertEqual(1, sa.vios_id)
        self.assertIsNotNone(sa.name)
        self.assertIsNotNone(sa.backing_dev_name)
        self.assertIsNotNone(sa.udid)
        self.assertEqual(sa.side, 'Server')
        self.assertTrue(sa.is_varied_on)
        self.assertIsNotNone(sa.lpar_slot_num)
        self.assertIsNotNone(sa.vios_slot_num)
        self.assertIsNotNone(sa.loc_code)

        # Try copying the map and adding it in
        new_map = copy.deepcopy(static_map)
        orig_size = len(mappings)
        mappings.append(new_map)
        self.assertEqual(len(mappings), orig_size + 1)
        self.assertEqual(len(self.dwrap.scsi_mappings), orig_size + 1)

        mappings.remove(new_map)
        self.dwrap.scsi_mappings = mappings
        self.assertEqual(len(self.dwrap.scsi_mappings), orig_size)

    def test_vfc_mappings(self):
        mappings = self.dwrap.vfc_mappings

        # Ensure that at least one adapter has a client LPAR
        found_client_uri = False
        static_map = None
        for mapping in mappings:
            if mapping.client_lpar_href:
                found_client_uri = True
                static_map = mapping
        self.assertTrue(found_client_uri)

        # We'll use the previous mapping as a baseline for further validation
        self.assertIsNotNone(static_map.client_adapter)
        self.assertIsNotNone(static_map.backing_port)
        self.assertIsNotNone(static_map.server_adapter)

        # Deeper check on each of these.
        ca = static_map.client_adapter
        self.assertIsNotNone(ca.wwpns)
        self.assertIsNotNone(ca.lpar_id)
        self.assertIsNotNone(ca.vios_id)
        self.assertTrue(ca.is_varied_on)
        self.assertIsNotNone(ca.lpar_slot_num)
        self.assertIsNotNone(ca.vios_slot_num)
        self.assertIsNotNone(ca.loc_code)
        self.assertEqual(ca.side, 'Client')

        bport = static_map.backing_port
        self.assertIsNotNone(bport.loc_code)
        self.assertIsNotNone(bport.name)
        self.assertIsNotNone(bport.udid)
        self.assertIsNotNone(bport.wwpn)
        self.assertIsNotNone(bport.npiv_available_ports)
        self.assertIsNotNone(bport.npiv_total_ports)

        sa = static_map.server_adapter
        self.assertIsNotNone(sa.name)
        self.assertIsNotNone(sa.map_port)
        self.assertIsNotNone(sa.udid)
        self.assertEqual(sa.side, 'Server')
        self.assertTrue(sa.is_varied_on)
        self.assertIsNotNone(sa.lpar_slot_num)
        self.assertIsNotNone(sa.vios_slot_num)
        self.assertIsNotNone(sa.loc_code)

        # Try copying the map and adding it in
        new_map = copy.deepcopy(static_map)
        orig_size = len(mappings)
        mappings.append(new_map)
        self.assertEqual(len(mappings), orig_size + 1)
        self.assertEqual(len(self.dwrap.vfc_mappings), orig_size + 1)

        mappings.remove(new_map)
        self.dwrap.vfc_mappings = mappings
        self.assertEqual(len(self.dwrap.vfc_mappings), orig_size)

    def test_bld_vfc_mapping(self):
        mapping = vios.VFCMapping.bld(self.adpt, 'host_uuid',
                                      'client_lpar_uuid', 'fcs0', ['aa', 'bb'])
        self.assertIsNotNone(mapping)

        # Validate the FC Backing port
        self.assertIsNotNone(mapping.backing_port)

        # Validate the Server Adapter
        self.assertIsNotNone(mapping.server_adapter)

        # Validate the Client Adapter
        self.assertIsNotNone(mapping.client_adapter)
        self.assertEqual(['AA', 'BB'], mapping.client_adapter.wwpns)

    def test_bld_vfc_mapping_with_slot(self):
        mapping = vios.VFCMapping.bld(self.adpt, 'host_uuid',
                                      'client_lpar_uuid', 'fcs0',
                                      client_wwpns=['aa', 'bb'],
                                      lpar_slot_num=3)
        self.assertIsNotNone(mapping)

        # Validate the FC Backing port
        self.assertIsNotNone(mapping.backing_port)

        # Validate the Server Adapter
        self.assertIsNotNone(mapping.server_adapter)

        # Validate the Client Adapter
        self.assertIsNotNone(mapping.client_adapter)
        self.assertEqual(['AA', 'BB'], mapping.client_adapter.wwpns)
        # verify the slot number
        self.assertEqual(3, mapping.client_adapter.lpar_slot_num)
        # Assert that we set this to False when specifying the slot number
        self.assertFalse(mapping.client_adapter._get_val_bool(
            'UseNextAvailableSlotID'))

    def test_bld_scsi_mapping_from_existing(self):
        def map_has_pieces(smap, lpar_href=True, client_adapter=True,
                           server_adapter=True, storage=True,
                           target_device=True):
            def has_piece(piece, has_it):
                if has_it:
                    self.assertIsNotNone(piece)
                else:
                    self.assertIsNone(piece)
            has_piece(smap.client_lpar_href, lpar_href)
            has_piece(smap.client_adapter, client_adapter)
            has_piece(smap.server_adapter, server_adapter)
            has_piece(smap.backing_storage, storage)
            has_piece(smap.element.find('TargetDevice'), target_device)
        stg = pvm_stor.VDisk.bld_ref(self.adpt, 'disk_name')
        smaps = self.dwrap.scsi_mappings
        # 0 has only ServerAdapter
        sm = smaps[0]
        map_has_pieces(sm, lpar_href=False, client_adapter=False,
                       storage=False, target_device=False)
        smclone = vios.VSCSIMapping.bld_from_existing(sm, stg)
        map_has_pieces(smclone, lpar_href=False, client_adapter=False,
                       target_device=False)
        self.assertEqual(stg, smclone.backing_storage)
        # 1 has ServerAdapter, Storage, and TargetDevice
        sm = smaps[1]
        map_has_pieces(sm, lpar_href=False, client_adapter=False)
        self.assertNotEqual(stg, sm.backing_storage)
        smclone = vios.VSCSIMapping.bld_from_existing(sm, stg)
        # Target device *disappears*
        map_has_pieces(smclone, lpar_href=False, client_adapter=False,
                       target_device=False)
        self.assertEqual(stg, smclone.backing_storage)
        # 3 has AssociatedLogicalPartition, ClientAdapter, ServerAdapter.
        sm = smaps[3]
        map_has_pieces(sm, storage=False, target_device=False)
        smclone = vios.VSCSIMapping.bld_from_existing(sm, stg)
        map_has_pieces(smclone, target_device=False)
        self.assertEqual(stg, smclone.backing_storage)
        # 12 has everything
        sm = smaps[12]
        map_has_pieces(sm)
        self.assertNotEqual(stg, sm.backing_storage)
        smclone = vios.VSCSIMapping.bld_from_existing(sm, stg)
        # Target device *disappears*
        map_has_pieces(smclone, target_device=False)
        self.assertEqual(stg, smclone.backing_storage)
        # Everything else cloned okay
        self.assertEqual(sm.client_lpar_href, smclone.client_lpar_href)
        self.assertEqual(sm.client_adapter, smclone.client_adapter)
        self.assertEqual(sm.server_adapter, smclone.server_adapter)


class TestCrtRelatedHref(unittest.TestCase):
    @mock.patch('pypowervm.adapter.Session')
    def test_crt_related_href(self, mock_sess):
        """Tests to make sure that related elements are well formed."""
        mock_sess.dest = 'root'
        adapter = adpt.Adapter(mock_sess)
        href = vios.VStorageMapping.crt_related_href(adapter, 'host', 'lpar')
        self.assertEqual('root/rest/api/uom/ManagedSystem/host/'
                         'LogicalPartition/lpar', href)


class TestVSCSIBus(twrap.TestWrapper):
    file = 'vscsibus_feed.txt'
    wrapper_class_to_test = vios.VSCSIBus

    def test_props(self):
        self.assertEqual(4, len(self.entries))
        bus = self.dwrap
        self.assertEqual('1f25efc1-a42b-3384-85e7-f37158f46615', bus.uuid)
        self.assertEqual(
            'http://localhost:12080/rest/api/uom/ManagedSystem/1cab7366-6b73-3'
            '42c-9f43-ddfeb9f8edd3/LogicalPartition/3DFF2EF5-6F99-4C29-B655-EE'
            '57DF1B64C6', bus.client_lpar_href)
        self.assertEqual(5, bus.client_adapter.lpar_id)
        self.assertEqual(2, bus.client_adapter.lpar_slot_num)
        self.assertEqual(5, bus.server_adapter.lpar_id)
        self.assertEqual(2, bus.server_adapter.lpar_slot_num)
        map1, map2 = bus.mappings
        self.assertIsInstance(map1.backing_storage, pvm_stor.PV)
        self.assertEqual('hdisk10', map1.backing_storage.name)
        self.assertIsInstance(map1.target_dev, pvm_stor.PVTargetDev)
        self.assertEqual('0x8100000000000000', map1.target_dev.lua)
        self.assertIsInstance(map2.backing_storage, pvm_stor.VOptMedia)
        self.assertEqual('cfg_My_OS_Image_V_3dff2ef5_000000.iso',
                         map2.backing_storage.name)
        self.assertIsInstance(map2.target_dev, pvm_stor.VOptTargetDev)
        self.assertEqual('0x8200000000000000', map2.target_dev.lua)

    def test_bld(self):
        self.adpt.build_href.return_value = 'href'
        # Default slot number (use next available)
        bus = vios.VSCSIBus.bld(self.adpt, 'client_lpar_uuid')
        self.adpt.build_href.assert_called_once_with(
            'LogicalPartition', 'client_lpar_uuid', xag=[])
        self.assertEqual('href', bus.client_lpar_href)
        self.assertTrue(bus.client_adapter._get_val_bool(
            pvm_stor._VADPT_NEXT_SLOT))
        self.assertIsNotNone(bus.server_adapter)
        self.assertEqual([], bus.mappings)
        # Specify slot number
        bus = vios.VSCSIBus.bld(self.adpt, 'client_lpar_uuid',
                                lpar_slot_num=42)
        self.assertFalse(bus.client_adapter._get_val_bool(
            pvm_stor._VADPT_NEXT_SLOT))
        self.assertEqual(42, bus.client_adapter.lpar_slot_num)

    def test_bld_from_existing(self):
        bus = vios.VSCSIBus.bld_from_existing(self.dwrap)
        self.assertEqual(
            'http://localhost:12080/rest/api/uom/ManagedSystem/1cab7366-6b73-3'
            '42c-9f43-ddfeb9f8edd3/LogicalPartition/3DFF2EF5-6F99-4C29-B655-EE'
            '57DF1B64C6', bus.client_lpar_href)
        self.assertEqual(5, bus.client_adapter.lpar_id)
        self.assertEqual(2, bus.client_adapter.lpar_slot_num)
        self.assertEqual(5, bus.server_adapter.lpar_id)
        self.assertEqual(2, bus.server_adapter.lpar_slot_num)
        self.assertEqual([], bus.mappings)

    def test_mappings(self):
        # No LUA
        lu1 = pvm_stor.LU.bld_ref(self.adpt, 'lu1', 'lu_udid1')
        std1 = vios.STDev.bld(self.adpt, lu1)
        self.assertIsInstance(std1.backing_storage, pvm_stor.LU)
        self.assertEqual('lu1', std1.backing_storage.name)
        self.assertIsNone(std1.target_dev)
        # With LUA
        vdisk1 = pvm_stor.VDisk.bld_ref(self.adpt, 'vdisk1')
        std2 = vios.STDev.bld(self.adpt, vdisk1, lua='vdisk1_lua')
        self.assertIsInstance(std2.backing_storage, pvm_stor.VDisk)
        self.assertEqual('vdisk1', std2.backing_storage.name)
        self.assertIsInstance(std2.target_dev, pvm_stor.VDiskTargetDev)
        self.assertEqual('vdisk1_lua', std2.target_dev.lua)
        # Add 'em to a bus
        bus = self.dwrap
        self.assertEqual(2, len(bus.mappings))
        bus.mappings.extend((std1, std2))
        self.assertEqual(4, len(bus.mappings))
        self.assertEqual('lu1', bus.mappings[2].backing_storage.name)
        self.assertEqual('vdisk1', bus.mappings[3].backing_storage.name)
        # Replace bus mappings
        bus.mappings = [std2, std1]
        self.assertEqual(2, len(bus.mappings))
        self.assertEqual('vdisk1', bus.mappings[0].backing_storage.name)
        self.assertEqual('lu1', bus.mappings[1].backing_storage.name)


class TestPartitionIOConfiguration(twrap.TestWrapper):

    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def setUp(self):
        super(TestPartitionIOConfiguration, self).setUp()
        self.io_config = self.dwrap.io_config

    def test_max_slots(self):
        self.assertEqual(80, self.io_config.max_virtual_slots)

    def test_io_slots(self):
        # IO Slots are typically associated with the VIOS.  Further testing
        # driven there.
        self.assertIsNotNone(self.io_config.io_slots)
        self.assertEqual(3, len(self.io_config.io_slots))


class TestIOSlots(twrap.TestWrapper):

    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def setUp(self):
        super(TestIOSlots, self).setUp()
        self.io_slot = self.dwrap.io_config.io_slots[0]

    def test_attrs(self):
        self.assertEqual('PCI-E SAS Controller', self.io_slot.description)
        self.assertEqual('U78AB.001.WZSJBM3', self.io_slot.phys_loc)
        self.assertEqual(825, self.io_slot.pc_adpt_id)
        self.assertEqual(260, self.io_slot.pci_class)
        self.assertEqual(825, self.io_slot.pci_dev_id)
        self.assertEqual(826, self.io_slot.pci_subsys_dev_id)
        self.assertEqual(4116, self.io_slot.pci_mfg_id)
        self.assertEqual(1, self.io_slot.pci_rev_id)
        self.assertEqual(4116, self.io_slot.pci_vendor_id)
        self.assertEqual(4116, self.io_slot.pci_subsys_vendor_id)
        self.assertEqual(553713674, self.io_slot.drc_index)
        self.assertEqual('U78AB.001.WZSJBM3-P1-T9',
                         self.io_slot.drc_name)
        self.assertEqual(False, self.io_slot.bus_grp_required)
        self.assertEqual(False, self.io_slot.required)

    def test_io_slots_setter(self):
        old_len = len(self.dwrap.io_config.io_slots)
        new_io_slots = self.dwrap.io_config.io_slots[:]
        deleted_slot = new_io_slots[1]
        del new_io_slots[1]
        self.dwrap.io_config.io_slots = new_io_slots
        self.assertEqual(old_len - 1, len(self.dwrap.io_config.io_slots))
        self.assertNotIn(deleted_slot, self.dwrap.io_config.io_slots)

    @mock.patch('warnings.warn')
    def test_io_adpt(self, mock_warn):
        self.assertEqual('553713674', self.io_slot.io_adapter.id)
        # Verify deprecation warning on IOSlot.adapter
        self.assertEqual('553713674', self.io_slot.adapter.id)
        mock_warn.assert_called_with(mock.ANY, DeprecationWarning)

    def test_bld(self):
        new_slot = bp.IOSlot.bld(self.adpt, True, 12345678)
        self.assertEqual(False, new_slot.required)
        self.assertEqual(True, new_slot.bus_grp_required)
        self.assertEqual(12345678, new_slot.drc_index)


class TestGenericIOAdapter(twrap.TestWrapper):

    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def setUp(self):
        super(TestGenericIOAdapter, self).setUp()
        self.io_adpt = self.dwrap.io_config.io_slots[0].io_adapter

    def test_attrs(self):
        self.assertEqual('553713674', self.io_adpt.id)
        self.assertEqual('PCI-E SAS Controller', self.io_adpt.description)
        self.assertEqual('U78AB.001.WZSJBM3-P1-T9',
                         self.io_adpt.dev_name)
        self.assertEqual('U78AB.001.WZSJBM3-P1-T9',
                         self.io_adpt.drc_name)
        self.assertEqual('T9', self.io_adpt.phys_loc_code)
        self.assertFalse(isinstance(self.io_adpt, bp.PhysFCAdapter))


class TestPhysFCAdapter(twrap.TestWrapper):

    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def setUp(self):
        super(TestPhysFCAdapter, self).setUp()
        self.io_adpt = self.dwrap.io_config.io_slots[2].io_adapter

    def test_attrs(self):
        desc = '8 Gigabit PCI Express Dual Port Fibre Channel Adapter'

        self.assertEqual('553714177', self.io_adpt.id)
        self.assertEqual(desc, self.io_adpt.description)
        self.assertEqual('U78AB.001.WZSJBM3-P1-C2', self.io_adpt.dev_name)
        self.assertEqual('U78AB.001.WZSJBM3-P1-C2',
                         self.io_adpt.drc_name)
        self.assertEqual('C2', self.io_adpt.phys_loc_code)
        self.assertIsInstance(self.io_adpt, bp.PhysFCAdapter)

    def test_fc_ports(self):
        self.assertEqual(2, len(self.io_adpt.fc_ports))


class TestPhysFCPort(twrap.TestWrapper):

    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def setUp(self):
        super(TestPhysFCPort, self).setUp()
        self.io_port1 = self.dwrap.io_config.io_slots[2].io_adapter.fc_ports[0]
        self.io_port2 = self.dwrap.io_config.io_slots[2].io_adapter.fc_ports[1]

    def test_attrs(self):
        self.assertEqual('U78AB.001.WZSJBM3-P1-C2-T2', self.io_port1.loc_code)
        self.assertEqual('fcs1', self.io_port1.name)
        self.assertEqual('1aU78AB.001.WZSJBM3-P1-C2-T2', self.io_port1.udid)
        self.assertEqual('10000090FA1B6303', self.io_port1.wwpn)
        self.assertEqual(0, self.io_port1.npiv_available_ports)
        self.assertEqual(0, self.io_port1.npiv_total_ports)

        self.assertEqual('U78AB.001.WZSJBM3-P1-C2-T1', self.io_port2.loc_code)
        self.assertEqual('fcs0', self.io_port2.name)
        self.assertEqual('1aU78AB.001.WZSJBM3-P1-C2-T1', self.io_port2.udid)
        self.assertEqual('10000090FA1B6302', self.io_port2.wwpn)
        self.assertEqual(64, self.io_port2.npiv_available_ports)
        self.assertEqual(64, self.io_port2.npiv_total_ports)


class TestIOAdapterChoices(twrap.TestWrapper):

    file = 'fake_vios_ssp_npiv.txt'
    wrapper_class_to_test = vios.VIOS

    def setUp(self):
        super(TestIOAdapterChoices, self).setUp()
        self.io_adpts = self.dwrap.io_adpts_for_link_agg

    def test_adapter_choices(self):
        self.assertEqual(len(self.io_adpts), 3)
        self.assertEqual(self.io_adpts[0].id, '1')
        self.assertEqual(
            self.io_adpts[0].description,
            '4-Port Gigabit Ethernet PCI-Express Adapter (e414571614102004)')
        self.assertEqual(self.io_adpts[0].dev_name, 'ent3')
        self.assertEqual(self.io_adpts[0].dev_type, 'physicalEthernetAdpter')
        self.assertEqual(self.io_adpts[0].drc_name,
                         'U78AB.001.WZSJBM3-P1-C7-T4')
        self.assertEqual(self.io_adpts[0].phys_loc_code,
                         'U78AB.001.WZSJBM3-P1-C7-T4')
        self.assertEqual(self.io_adpts[0].udid,
                         '13U78AB.001.WZSJBM3-P1-C7-T4')


class TestFeed3(twrap.TestWrapper):
    """Tests that specifically need fake_vios_feed3.txt"""
    file = 'fake_vios_feed3.txt'
    wrapper_class_to_test = vios.VIOS

    def test_vivify_io_adpts_for_link_agg(self):
        """Vivifying FreeIOAdaptersForLinkAggregation adds the Network xag."""
        # The first VIOS doesn't have FreeIOAdapters...
        vwrp = self.dwrap
        self.assertIsNone(vwrp._find(vios._VIO_FREE_IO_ADPTS_FOR_LNAGG))
        # Vivify it - should be empty
        self.assertEqual([], vwrp.io_adpts_for_link_agg)
        # Now it's in there
        elem = vwrp._find(vios._VIO_FREE_IO_ADPTS_FOR_LNAGG)
        self.assertIsNotNone(elem)
        # Got the right xag
        self.assertEqual(c.XAG.VIO_NET, elem.attrib['group'])

    @mock.patch('warnings.warn')
    def test_xags(self, mock_warn):
        """Test deprecated extented attribute groups on the VIOS class.

        This can be removed once VIOS.xags is removed.
        """
        expected = dict(NETWORK=c.XAG.VIO_NET, STORAGE=c.XAG.VIO_STOR,
                        SCSI_MAPPING=c.XAG.VIO_SMAP, FC_MAPPING=c.XAG.VIO_FMAP)

        for key, val in expected.items():
            # Test class accessor, ensure '.name' works.
            self.assertEqual(val, getattr(vios.VIOS.xags, key).name)
            mock_warn.assert_called_with(mock.ANY, DeprecationWarning)
            mock_warn.reset_mock()
            # Test instance accessor.
            self.assertEqual(val, getattr(self.dwrap.xags, key))
            mock_warn.assert_called_with(mock.ANY, DeprecationWarning)
            mock_warn.reset_mock()

        # And in case getattr(foo, 'bar') actually differs from foo.bar...
        self.assertEqual(c.XAG.VIO_NET, vios.VIOS.xags.NETWORK)
        mock_warn.assert_called_with(mock.ANY, DeprecationWarning)
        mock_warn.reset_mock()
        # Make sure the equality comparison works the other way
        self.assertEqual(self.dwrap.xags.NETWORK, c.XAG.VIO_NET)
        mock_warn.assert_called_with(mock.ANY, DeprecationWarning)
        # Test sorting
        self.assertTrue(c.XAG.VIO_NET < self.dwrap.xags.SCSI_MAPPING)
        self.assertTrue(self.dwrap.xags.NETWORK < c.XAG.VIO_SMAP)

if __name__ == "__main__":
    unittest.main()



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

import pypowervm.entities as ent
from pypowervm.tasks import migration as mig
import pypowervm.tests.tasks.util as u
import pypowervm.tests.test_fixtures as fx


class TestMigration(testtools.TestCase):
    """Unit Tests for Migration."""

    def setUp(self):
        super(TestMigration, self).setUp()
        self.adpt = self.useFixture(fx.AdapterFx()).adpt
        mock_resp = mock.MagicMock()
        mock_resp.entry = ent.Entry(
            {}, ent.Element('Dummy', self.adpt), self.adpt)
        self.adpt.read.return_value = mock_resp
        self.lpar_w = mock.MagicMock()
        self.lpar_w.adapter = self.adpt
        self.lpar_w.uuid = '1234'

    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    def test_migration(self, mock_run_job):

        # Test simple call
        mock_run_job.side_effect = u.get_parm_checker(
            self, '1234', [(mig.TGT_MGD_SYS, 'abc')], exp_timeout=1800 * 4)
        mig.migrate_lpar(self.lpar_w, 'abc')
        self.adpt.read.assert_called_once_with('LogicalPartition', '1234',
                                               suffix_parm='Migrate',
                                               suffix_type='do')
        # Test all parms
        self.adpt.read.reset_mock()
        parm_list = [(mig.TGT_MGD_SYS, 'abc'),
                     (mig.TGT_RMT_HMC, 'host'),
                     (mig.TGT_RMT_HMC_USR, 'usr'),
                     (mig.DEST_MSP, 'vios1'),
                     (mig.SRC_MSP, 'vios2'),
                     (mig.SPP_ID, '5'),
                     (mig.OVS_OVERRIDE, '2'),
                     (mig.VLAN_BRIDGE_OVERRIDE, '2')]
        mapping_list = [(mig.VFC_MAPPINGS, ['1/1/1', '3/3/3//3']),
                        (mig.VSCSI_MAPPINGS, ['2/2/2']),
                        (mig.VLAN_MAPPINGS, ['001122334455/4',
                                             '001122334466/5/6 7'])]
        mock_run_job.side_effect = u.get_parm_checker(
            self, '1234', parm_list, exp_job_mappings=mapping_list,
            exp_timeout=1800 * 4)

        mig.migrate_lpar(self.lpar_w, 'abc',
                         tgt_mgmt_svr='host', tgt_mgmt_usr='usr',
                         virtual_fc_mappings=['1/1/1', '3/3/3//3'],
                         virtual_scsi_mappings=['2/2/2'],
                         vlan_mappings=['001122334455/4',
                                        '001122334466/5/6 7'],
                         dest_msp_name='vios1', source_msp_name='vios2',
                         spp_id='5', sdn_override=True,
                         vlan_check_override=True)
        self.adpt.read.assert_called_once_with('LogicalPartition', '1234',
                                               suffix_parm='Migrate',
                                               suffix_type='do')
        # Test simple validation call
        self.adpt.read.reset_mock()
        mock_run_job.side_effect = u.get_parm_checker(
            self, '1234', [(mig.TGT_MGD_SYS, 'abc')], exp_timeout=1800 * 4)
        mock_run_job.reset_mock()
        mig.migrate_lpar(self.lpar_w, 'abc', validate_only=True)
        self.adpt.read.assert_called_once_with('LogicalPartition', '1234',
                                               suffix_parm='MigrateValidate',
                                               suffix_type='do')

    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    def test_migration_recover(self, mock_run_job):
        # Test simple call
        mig.migrate_recover(self.lpar_w)
        self.adpt.read.assert_called_once_with('LogicalPartition', '1234',
                                               suffix_parm='MigrateRecover',
                                               suffix_type='do')
        mock_run_job.assert_called_once_with(
            '1234', job_parms=[], timeout=1800)

        # Test simple call with force
        self.adpt.read.reset_mock()
        mock_run_job.reset_mock()
        mock_run_job.side_effect = u.get_parm_checker(
            self, '1234', [('Force', 'true')], exp_timeout=1800)
        mig.migrate_recover(self.lpar_w, force=True)

        self.adpt.read.assert_called_once_with('LogicalPartition', '1234',
                                               suffix_parm='MigrateRecover',
                                               suffix_type='do')

    @mock.patch('pypowervm.wrappers.job.Job.run_job')
    def test_migration_abort(self, mock_run_job):
        # Test simple call
        mig.migrate_abort(self.lpar_w)
        self.adpt.read.assert_called_once_with('LogicalPartition', '1234',
                                               suffix_parm='MigrateAbort',
                                               suffix_type='do')
        mock_run_job.assert_called_once_with(
            '1234', job_parms=None, timeout=1800)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""EntryWrapper for File ('web' namespace)."""

import pypowervm.const as pc
import pypowervm.wrappers.entry_wrapper as ewrap

_FILE_NAME = 'Filename'
_FILE_DATE_MOD = 'DateModified'
_FILE_INET_MED_TYPE = 'InternetMediaType'
_FILE_UUID = 'FileUUID'
_FILE_EXP_SIZE = 'ExpectedFileSizeInBytes'
_FILE_CUR_SIZE = 'CurrentFileSizeInBytes'
_FILE_ENUM_TYPE = 'FileEnumType'
_FILE_VIOS = 'TargetVirtualIOServerUUID'
_FILE_TDEV_UDID = 'TargetDeviceUniqueDeviceID'
_FILE_ASSET_FILE = 'AssetFile'
_FILE_CHKSUM = 'SHA256'


_DEFAULT_MEDIA_TYPE = 'application/octet-stream'


class FileType(object):
    """Supported file types."""
    MEDIA_ISO = 'BROKERED_MEDIA_ISO'
    DISK_IMAGE = 'BROKERED_DISK_IMAGE'
    # Obsolete.  Behaves the same as DISK_IMAGE.
    DISK_IMAGE_COORDINATED = 'BROKERED_DISK_IMAGE'


@ewrap.EntryWrapper.pvm_type('File', ns=pc.WEB_NS)
class File(ewrap.EntryWrapper):
    """Wraps the File Metadata for files on the VIOS.

    The API supports passing a File up to devices on the Virtual I/O Server.
    This object wraps the metadata for the Files.
    """

    @classmethod
    def bld(cls, adapter, f_name, f_type, v_uuid, sha_chksum=None, f_size=None,
            tdev_udid=None):
        """Creates a fresh File wrapper that can be used for a create action.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param f_name: The name for the file.
        :param f_type: The type of the file.  One of the FileType values.
        :param v_uuid: The UUID for the Virtual I/O Server that the file will
                       reside on.
        :param sha_chksum: (OPTIONAL) The SHA256 checksum for the file.  Useful
                           for integrity checks.
        :param f_size: (OPTIONAL) The size in bytes of the file to upload.  Can
                       be an int or a String (that represents an integer
                       number).  Useful for integrity checks.
        :param tdev_udid: The device UDID that the file will back into.
        :returns: The newly created File wrapper.
        """
        # Metadata needs to be in a specific order.  These are required
        f = super(File, cls)._bld(adapter)
        f._file_name(f_name)
        f._internet_media_type(_DEFAULT_MEDIA_TYPE)

        # Optional - should not be included in the Element if None.
        if sha_chksum:
            f._chksum(sha_chksum)

        if f_size:
            f._expected_file_size(f_size)

        # These are required
        f._enum_type(f_type)
        f._vios_uuid(v_uuid)

        # Optical media doesn't need to cite a target dev for file upload
        if tdev_udid:
            f._tdev_udid(tdev_udid)

        return f

    @property
    def file_name(self):
        return self._get_val_str(_FILE_NAME)

    def _file_name(self, name):
        self.set_parm_value(_FILE_NAME, name)

    @property
    def date_modified(self):
        return self._get_val_str(_FILE_DATE_MOD)

    @property
    def internet_media_type(self):
        """Typically 'application/octet-stream'."""
        return self._get_val_str(_FILE_INET_MED_TYPE)

    def _internet_media_type(self, imt):
        self.set_parm_value(_FILE_INET_MED_TYPE, imt)

    @property
    def file_uuid(self):
        """The file's UUID (different from the entries)."""
        return self._get_val_str(_FILE_UUID)

    @property
    def expected_file_size(self):
        return self._get_val_int(_FILE_EXP_SIZE)

    def _expected_file_size(self, sz):
        self.set_parm_value(_FILE_EXP_SIZE, sz)

    @property
    def current_file_size(self):
        return self._get_val_int(_FILE_CUR_SIZE)

    @property
    def enum_type(self):
        """The type of the file.  One of the FileType values."""
        return self._get_val_str(_FILE_ENUM_TYPE)

    def _enum_type(self, et):
        self.set_parm_value(_FILE_ENUM_TYPE, et)

    def _chksum(self, sha):
        self.set_parm_value(_FILE_CHKSUM, sha)

    @property
    def vios_uuid(self):
        return self._get_val_str(_FILE_VIOS)

    def _vios_uuid(self, uuid):
        self.set_parm_value(_FILE_VIOS, uuid)

    @property
    def tdev_udid(self):
        return self._get_val_str(_FILE_TDEV_UDID)

    def _tdev_udid(self, udid):
        self.set_parm_value(_FILE_TDEV_UDID, udid)

    @property
    def asset_file(self):
        """Used to identify the asset file on upload.

        Only used in conjunction with DISK_IMAGE_COORDINATED.

        Provides the path to a file on the local system where data can be sent
        during an upload operation.  This is used for significant speed
        improvements as the REST API server does not need to be involved with
        the upload.
        """
        return self._get_val_str(_FILE_ASSET_FILE)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

import mock
import testtools

import pypowervm.adapter as adp
import pypowervm.exceptions as pvmex
import pypowervm.helpers.sample_helper as smpl_hlp
import pypowervm.tests.test_fixtures as fx


class TestSampleHelper(testtools.TestCase):

    def setUp(self):
        super(TestSampleHelper, self).setUp()
        self.sess = self.useFixture(fx.SessionFx()).sess

    @mock.patch('time.sleep')
    def test_sample_helper(self, mock_sleep):

        helpers = smpl_hlp.sample_retry_helper
        fake_resp1 = adp.Response(
            'GET', '/some/path', 200, 'OK', ['headers'],
            body='Some Text HSCL3205 More Text')
        self.sess.request.side_effect = pvmex.Error('yo', response=fake_resp1)
        adpt = adp.Adapter(self.sess, helpers=helpers)
        self.assertRaises(
            pvmex.Error, adpt._request, 'method', 'path', body='the body')

        # Test that the request method was called twice and sleep was called
        self.assertEqual(self.sess.request.call_count, 2)
        mock_sleep.assert_called_once_with(5 * 1)

        hlp = functools.partial(smpl_hlp.sample_retry_helper, max_retries=5)
        self.sess.reset_mock()
        try:
            adpt._request('method', 'path', body='the body', helpers=hlp)
        except Exception:
            # Should have tried 6 times total
            self.assertEqual(self.sess.request.call_count, 6)



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""This Adapter helper logs recent requests/responses on an exception."""

import collections
import copy
import threading

from oslo_log import log as logging

import pypowervm.const as c
import pypowervm.exceptions as pvmex
from pypowervm.i18n import _


LOG = logging.getLogger(__name__)

log_store = threading.local()


def _init_thread_stg(max_entries):
    """Sets up the storage for the logs for this thread."""
    if not hasattr(log_store, 'powervm_log'):
        log_store.powervm_log = collections.deque(maxlen=max_entries)


def _stash(sensitive, type_, value):
    """Enters the request or response in the thread log."""
    if sensitive:
        value = '<SENSITIVE>'
    log_store.powervm_log.append({type_: value})


def _stash_response(sensitive, resp):
    if resp is not None:
        logged_resp = dict(resp.__dict__)
        # Remove feed and entry from the dictionary of response attributes.
        # This avoids keeping heavy ElementTree objects in memory since
        # only the string version of the body is dumped for responses.
        logged_resp.pop('entry', None)
        logged_resp.pop('feed', None)
        _stash(sensitive, 'response', logged_resp)


def _write_thread_log():
    def format_request(req):
        body = None
        # Parse the arguments if we're passed a tuple else its a string
        if isinstance(req, tuple):
            req_args = req[0]
            req_kwds = req[1]
            dump = dict(method=req_args[0], path=req_args[1])
            for key in req_kwds:
                if key == 'body':
                    # special format for body
                    body = req_kwds.get('body')
                elif key == 'headers':
                    # deep copy the header and change what we can't dump
                    headers = copy.deepcopy(req_kwds.get(key))
                    if 'X-API-Session' in headers:
                        headers['X-API-Session'] = '<SENSITIVE>'
                    dump[key] = str(headers)
                else:
                    dump[key] = str(req_kwds.get(key))
        else:
            dump = req

        # Dump all fields besides the body
        LOG.info(_('REQUEST: %s') % dump)
        # Now dump the full body
        if body is not None:
            LOG.info(body)

    def format_response(resp):
        body = None
        # Parse the arguments if we're passed a dict else it's a string
        if isinstance(resp, dict):
            dump = {}
            for key in resp:
                if key == 'body':
                    # special format for body
                    body = resp.get('body')
                else:
                    dump[key] = str(resp.get(key))
        else:
            dump = resp

        # Dump all fields besides the body first
        LOG.info(_('RESPONSE: %s') % dump)
        # Now dump the full body, on the next line, if available
        if body is not None:
            LOG.info(body)

    # Pop each entry out of the log until it's empty
    try:
        while True:
            entry = log_store.powervm_log.popleft()
            request = entry.get('request')
            if request is not None:
                format_request(request)

            response = entry.get('response')
            if response is not None:
                format_response(response)
    except IndexError:
        pass


def log_helper(func, max_logs=3):
    """Log recent requests/responses on exception.

    This helper stashes the requests/responses it sees passing through to
    thread local storage.  If it then sees an exception surfacing, it will
    write the req/resp logs.

    :param func: The Adapter request method to call
    :param max_logs (int): Max number of req/resps to retain at a time
            This value can only be set once per thread.
            Once it's set, subsequent calls will ignore the value.
    """
    def is_etag_mismatch(ex):
        """Is ex an HttpError with status 412 (etag mismatch)?"""
        return (isinstance(ex, pvmex.HttpError) and
                ex.response.status == c.HTTPStatus.ETAG_MISMATCH)

    def log_req_resp(*args, **kwds):
        # Set aside storage for a req/resp pair
        _init_thread_stg(max_entries=(max_logs * 2))

        # See if this request has sensitive data
        sensitive = kwds.get('sensitive', False)
        # Log the request before the call
        _stash(sensitive, 'request', (args, kwds))
        try:
            # Call the request()
            response = func(*args, **kwds)
        except pvmex.Error as e:
            _stash_response(sensitive, e.response)
            # Now dump the log and raise the exception.
            # Special case for 412 (etag mismatch) - don't dump.
            if not is_etag_mismatch(e):
                _write_thread_log()
            raise
        else:
            _stash_response(sensitive, response)
            return response

    return log_req_resp



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
from concurrent.futures import thread as th
import oslo_concurrency.lockutils as lock
import oslo_context.context as ctx
from oslo_log import log as logging
from oslo_utils import reflection
import six
from taskflow import engines as tf_eng
from taskflow import exceptions as tf_ex
from taskflow.patterns import linear_flow as tf_lf
from taskflow.patterns import unordered_flow as tf_uf
from taskflow import task as tf_task
import threading

import pypowervm.exceptions as ex
from pypowervm.i18n import _
from pypowervm.utils import retry
import pypowervm.wrappers.entry_wrapper as ewrap

LOG = logging.getLogger(__name__)
_local = threading.local()


def _get_locks():
    """Returns the list of UUIDs locked by this thread."""
    locks = getattr(_local, 'entry_transaction', None)
    if locks is None:
        locks = []
        _set_locks(locks)
    return locks


def _set_locks(locks):
    """Sets the list of UUIDs locked by this thread."""
    _local.entry_transaction = locks


def entry_transaction(func):
    """Decorator to facilitate transaction semantics on a PowerVM object.

    Typically, a method thus decorated will make some set of changes to an
    EntryWrapper and then perform one or more REST operations thereon.

    The *consumer* of the decorated method may pass either an EntryWrapper or
    an EntryWrapperGetter as the first argument.  The *developer* of the
    decorated method is guaranteed that the first argument is an EntryWrapper.

    This decorator provides three things:
    1) The decorated method may be invoked with either an EntryWrapper or an
    EntryWrapperGetter as its first argument.  However, within the body of the
    method, that argument is guaranteed to be the appropriate EntryWrapper.
    2) The decorated method is locked on the UUID of the PowerVM object on
    which it operates (represented by its first argument).  Only one method
    thus decorated can operate on that PowerVM object at one time.
    3) If the decorated method fails due to an etag mismatch - indicating that
    the wrapper was updated out-of-band between when it was retrieved and when
    it was updated - the wrapper is refreshed and the entire method is
    redriven.

    Note: If the etag mistmatch occurs, the STEPPED_DELAY function is used
    from the retry.py.  This provides a gradual increase in the delay (except
    for the first retry - which is immediate).  A maximum number of 6 retries
    will occur.

    Example usage:

    @entry_transaction
    def add_gizmos_to_vios_wrapper(vios_wrapper, gizmos):
        vios_wrapper.gizmo_list.extend(gizmos)
        return vios_wrapper.update()

    This method can then be invoked either as:

    add_gizmos_to_vios_wrapper(existing_vios_wrapper, gizmos)

    or as:

    add_gizmos_to_vios_wrapper(pvm_vios.VIOS.getter(adapter, uuid), gizmos)
    """
    def _synchronize(wrp_or_spec, *a1, **k1):
        """Returned method is synchronized on the object's UUID."""

        @lock.synchronized(wrp_or_spec.uuid)
        def _locked_resolve_wrapper(wos, *a2, **k2):
            try:
                # The synchronized decorator will hold off other threads
                # we just have to hold off lock attempts by methods further
                # down the stack.
                _get_locks().append(wrp_or_spec.uuid)
                return _resolve_wrapper(wos, *a2, **k2)
            finally:
                _get_locks().remove(wrp_or_spec.uuid)

        def _resolve_wrapper(wos, *a2, **k2):
            """Returned method guaranteed to be called with a wrapper."""
            if isinstance(wos, ewrap.EntryWrapperGetter):
                wos = wos.get()

            @retry.retry(argmod_func=retry.refresh_wrapper, tries=60,
                         delay_func=retry.STEPPED_RANDOM_DELAY)
            def _retry_refresh(wrapper, *a3, **k3):
                """Retry as needed, refreshing its wrapper each time."""
                return func(wrapper, *a3, **k3)
            return _retry_refresh(wos, *a2, **k2)

        def _lock_if_needed(wos, *a2, **k2):
            # Check if this UUID is already locked
            if wrp_or_spec.uuid in _get_locks():
                # It's already locked by this thread, so skip the lock.
                return _resolve_wrapper(wos, *a2, **k2)
            else:
                return _locked_resolve_wrapper(wos, *a2, **k2)

        return _lock_if_needed(wrp_or_spec, *a1, **k1)
    return _synchronize


@six.add_metaclass(abc.ABCMeta)
class Subtask(object):
    """A single EntryWrapper modification to be performed within a WrapperTask.

    A subclass performs its work by overriding the execute method.  That method
    may or may not make changes to the EntryWrapper, which is its first
    argument.  Its return value must indicate whether changes were made to the
    wrapper: this is the trigger used by WrapperTask to determine whether to
    POST the changes back to the REST server via update().  The return value is
    saved by the surrounding WrapperTask if the 'provides' argument is used on
    initialization.  This value can then be retrieved by subsequent Subtasks.

    A Subtask should never update() or refresh() the wrapper.  That is handled
    by the surrounding WrapperTask.

    See WrapperTask for example usage.
    """
    def __init__(self, *save_args, **save_kwargs):
        """Create the Subtask, saving execution arguments for later.

        :param save_args: Positional arguments to be passed to the execute
                          method - *after* the wrapper - when it is invoked
                          under a WrapperTask.
        :param save_kwargs: Keyword arguments to be passed to the execute
                            method when it is invoked under a WrapperTask.
        :param provides: (Optional) String name for the return value from the
                         execute method.  If this parameter is used, the return
                         value will be saved by the surrounding WrapperTask and
                         be available to subsequent Subtasks via the 'provided'
                         keyword argument.  The 'provides' name must be unique
                         within a WrapperTask.
        :param flag_update: (Optional) Boolean indicating whether a True return
                            from this Subtask should trigger an update() in the
                            surrounding WrapperTask.  By default, this is True.
                            Set this to False, for example, to provide some
                            data to subsequent Subtasks without forcing an
                            update.
        """
        self.provides = save_kwargs.pop('provides', None)
        self.flag_update = save_kwargs.pop('flag_update', True)
        self.save_args = save_args
        self.save_kwargs = save_kwargs

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        """Modify the EntryWrapper (must be overridden by the subclass).

        The execute method has two responsibilities:
        1) Performs the modification to the EntryWrapper which is passed as its
        first argument.
        2) Indicates whether any modifications were performed.

        Example:
        def execute(thingy_wrapper, primary_widget, provided=None):
            update_needed = False
            if primary_widget not in thingy_wrapper.widgets:
                thingy_wrapper.set_primary_widget(primary_widget)
                update_needed = True
            # Was a widget list provided by a prior Subtask?
            if provided is not None:
                widget_list = provided.get('widget_list', [])
                for widget in widget_list:
                    thingy_wrapper.widgets.append(widget)
                    update_needed = True
            return update_needed

        :param args: Positional arguments accepted by the execute method.  The
                     first argument will always be the EntryWrapper.  Overrides
                     may define their signatures using explicit parameter
                     names.
        :param kwargs: Keyword arguments accepted by the execute method.
                       Overrides may use explicit parameter names.
        :param provided: Dict of return values provided by Subtasks whose
                         execution preceded this one, and which used the
                         'provides' keyword argument to save their returns.
                         The keys of the dict are the 'provides' strings of the
                         prior Subtasks.
        :return: The return value must be a single value (this may be a list,
                 but not a tuple) which evaluates to True or False.  Unless
                 this Subtask was initialized with flag_update=False, any True
                 value indicates that the wrapper was modified and should be
                 POSTed back to the REST server via update().  Any False value
                 (including None, [], {}, etc) indicates that this Subtask did
                 not modify the wrapper.  (Note that it may still be POSTed if
                 modified by other Subtasks in the same WrapperTask.)
        """


class _FunctorSubtask(Subtask):
    """Shim to create a Subtask around an existing callable."""
    def __init__(self, _func, *save_args, **save_kwargs):
        """Save the callable as well as the arguments.

        :param _func: Callable to be invoked under the WrapperTask.
        :param save_args: See Subtask.__init__(save_args).
        :param save_kwargs: See Subtask.__init__(save_kwargs).  May contain the
                            following values, which are treated specially and
                            NOT passed to the callable _func:
               provides: See Subtask.__init__(provides).
               flag_update: See Subtask.__init__(flag_update).
               logspec: Iterable comprising a logging function, a format
                        string, and zero or more arguments.  The log method is
                        invoked before the func.  Example:

                    logspec = [LOG.info, _LI("Deleting widget %(widget)s from "
                                             "instance %(instance)s."),
                               {'widget': widg, 'instance': instance.name}]
                    FunctorSubtask(..., logspec=logspec)
        """
        self._logspec = save_kwargs.pop('logspec', [])
        super(_FunctorSubtask, self).__init__(*save_args, **save_kwargs)
        self._func = _func
        if self._logspec:
            if len(self._logspec) < 2 or not callable(self._logspec[0]):
                raise ValueError(
                    "logspec must be a list comprising a callable followed by "
                    "a format string and zero or more arguments.")

    def execute(self, wrapper, *_args, **_kwargs):
        """Invoke saved callable with saved args."""
        if not ('provided' in reflection.get_callable_args(self._func)
                or reflection.accepts_kwargs(self._func)):
            _kwargs.pop('provided', None)
        if self._logspec:
            # Execute the log method (the first element in the list) with its
            # arguments (the remaining elements in the list).
            self._logspec[0](*self._logspec[1:])
        return self._func(wrapper, *_args, **_kwargs)


class WrapperTask(tf_task.Task):
    """An atomic modify-and-POST transaction Task over a single EntryWrapper.

    The modifications should comprise some number of Subtask instances, added
    to this WrapperTask via the add_subtask and/or add_functor_subtask methods.
    These Subtasks should only modify the EntryWrapper, and should not POST
    (.update()) it back to the REST Server.  The WrapperTask will decide
    whether a POST is needed based on the returns from the Subtasks' execute
    methods, and perform it if indicated.

    The WrapperTask's execute method is encompassed by @entry_transaction,
    meaning that:
    1) The initial GET of the EntryWrapper may be deferred until after the lock
    is acquired.
    2) The execute method is locked on the UUID of the Entry in question.
    3) If the final update (POST) fails due to etag mismatch, the EntryWrapper
    is refetched and the entire transaction is redriven from the start.

    Usage:
        class ModifyGizmos(Subtask):
            def execute(self, wrapper, gizmo_list, provides='first_gizmo'):
                update_needed = None
                if gizmo_list:
                    wrapper.gizmos.append(gizmo_list)
                    update_needed = gizmo_list[0]
                return update_needed

        def add_widget(wrapper, widget, frob=False, provided=None):
            if provided is not None:
                widget.first_gizmo = provided.get('first_gizmo')
            wrapper.widgets.append(widget, frob)
            return len(wrapper.widgets)
        ...
        tx = WrapperTask("do_lpar_things", LPAR.getter(adapter, lpar_uuid))
      or
        tx = WrapperTask("do_lpar_things", LPAR.getter(adapter, lpar_uuid),
                         subtasks=existing_wrapper_task.subtasks)
      or
        # Not recommended - increased probability of retry
        wrapper = LPAR.wrap(adapter.read(LPAR.schema_type, lpar_uuid))
        tx = WrapperTask("do_lpar_things", wrapper)
        ...
        tx.add_subtask(ModifyGizmos([giz1, giz2]))
        ...
        logspec = [LOG.info, _LI("Added widget %(widget)s to LPAR %(lpar)s."),
                   {'widget': widget.name, 'lpar': lpar_uuid}]
        tx.add_functor_subtask(add_widget, widget, provides='widget_count',
                               logspec=logspec)
        ...
        finalized_lpar = tx.execute()
    """
    def __init__(self, name, wrapper_or_getter, subtasks=None,
                 allow_empty=False, update_timeout=-1):
        """Initialize this WrapperTask.

        :param name: A descriptive string name for the WrapperTask.
        :param wrapper_or_getter: An EntryWrapper or EntryWrapperGetter
                                  representing the PowerVM object on which this
                                  WrapperTask is to be performed.
        :param subtasks: (Optional) Iterable of Subtask subclass instances with
                         which to seed this WrapperTask.
        :param allow_empty: (Optional) By default, executing a WrapperTask
                            containing no Subtasks will result in exception
                            WrapperTaskNoSubtasks.  If this flag is set to
                            True, this condition will instead log an info
                            message and return None (NOT the wrapper - note,
                            this is different from "subtasks ran, but didn't
                            change anything," which returns the wrapper).
        :param update_timeout: (Optional) Integer number of seconds after which
                               to time out the POST request.  -1, the default,
                               causes the request to use the timeout value
                               configured on the Session belonging to the
                               Adapter.
        :raise WrapperTaskNoSubtasks: If allow_empty is False and this
                                      WrapperTask is executed without any
                                      Subtasks having been added.
        """
        if isinstance(wrapper_or_getter, ewrap.EntryWrapperGetter):
            self._wrapper = None
            self._getter = wrapper_or_getter
        elif isinstance(wrapper_or_getter, ewrap.EntryWrapper):
            self._wrapper = wrapper_or_getter
            self._getter = None
        else:
            raise ValueError(_("Must supply either EntryWrapper or "
                               "EntryWrapperGetter."))
        super(WrapperTask, self).__init__(
            name, provides=('wrapper_%s' % wrapper_or_getter.uuid,
                            'subtask_rets_%s' % wrapper_or_getter.uuid))
        self._tasks = [] if subtasks is None else list(subtasks)
        self.allow_empty = allow_empty
        self.update_timeout = update_timeout
        # Dict of return values provided by Subtasks using the 'provides' arg.
        self.provided = {}
        # Set of 'provided' names to prevent duplicates.  (Some day we may want
        # to make this a list and use it to denote the order in which subtasks
        # were run.)
        self.provided_keys = set()

    def add_subtask(self, task):
        """Add a Subtask to this WrapperTask.

        Subtasks will be invoked serially and synchronously in the order in
        which they are added.

        :param task: Instance of a Subtask subclass containing the logic to
                     invoke.
        :return: self, for chaining convenience.
        """
        if not isinstance(task, Subtask):
            raise ValueError(_("Must supply a valid Subtask."))
        # Seed the 'provided' dict and ensure no duplicate names
        if task.provides is not None:
            if task.provides in self.provided_keys:
                raise ValueError(_("Duplicate 'provides' name %s.") %
                                 task.provides)
            self.provided_keys.add(task.provides)
        self._tasks.append(task)
        return self

    def add_functor_subtask(self, func, *args, **kwargs):
        """Create and add a Subtask for an already-defined method.

        :param func: A callable to be the core of the Subtask.  The contract
                     for this method is identical to that of Subtask.execute -
                     see that method's docstring for details.
        :param args: Positional arguments to be passed to the callable func
                     (after the EntryWrapper parameter) when it is executed
                     within the WrapperTask.
        :param kwargs: Keyword arguments to be passed to the callable func when
                       it is executed within the WrapperTask.  May contain the
                       following values, which are treated specially and NOT
                       passed to the callable func:
               provides: See Subtask.__init__(provides).
               flag_update: See Subtask.__init__(flag_update).
               logspec: Iterable comprising a logging function, a format
                        string, and zero or more arguments.  The log method is
                        invoked before the func.  Example:

                    logspec = [LOG.info, _LI("Deleting widget %(widget)s from "
                                             "instance %(instance)s."),
                               {'widget': widg, 'instance': instance.name}]
                    FunctorSubtask(..., logspec=logspec)
        :return: self, for chaining convenience.
        """
        return self.add_subtask(_FunctorSubtask(func, *args, **kwargs))

    @property
    def wrapper(self):
        """(Fetches and) returns the EntryWrapper.

        Use this only if you need the EntryWrapper outside of the WrapperTask's
        execution itself.

        Note that this guarantees a GET outside of lock, and should therefore
        be used only if absolutely necessary.
        """
        if not self._wrapper:
            self._wrapper = self._getter.get()
        # NOTE: This access of self._wrapper must remain atomic.
        # See TAG_WRAPPER_SYNC.
        return self._wrapper

    @property
    def subtasks(self):
        """Return the sequence of Subtasks registered with this WrapperTask.

        This is returned as a tuple (not modifiable).  To add subtasks, use the
        add_[functor_]subtask method.
        """
        return tuple(self._tasks)

    def execute(self):
        """Invoke subtasks and update under @entry_transaction.

        The flow is as follows:

        1 Lock on wrapper UUID
        2 GET wrapper if necessary
        3 For each registered Subtask:
            - Invoke the Subtask to modify the wrapper
        4 If update is necessary, POST the wrapper.  If POST fails with etag
          mismatch:
            - Refresh the wrapper
            - goto 2
        5 Unlock
        """
        if len(self._tasks) == 0:
            if self.allow_empty:
                LOG.info(_("WrapperTask %s has no Subtasks; no-op execution."),
                         self.name)
                return None
            raise ex.WrapperTaskNoSubtasks(name=self.name)

        @entry_transaction
        def _execute(wrapper):
            update_needed = False
            for task in self._tasks:
                kwargs = task.save_kwargs
                if ('provided' in reflection.get_callable_args(task.execute)
                        or reflection.accepts_kwargs(task.execute)):
                    kwargs['provided'] = self.provided
                ret = task.execute(wrapper, *task.save_args, **kwargs)
                if task.flag_update and ret:
                    update_needed = True
                if task.provides is not None:
                    self.provided[task.provides] = ret
            if update_needed:
                wrapper = wrapper.update(timeout=self.update_timeout)
            return wrapper
        # Use the wrapper if already fetched, or the getter if not
        # NOTE: This assignment must remain atomic.  See TAG_WRAPPER_SYNC.
        self._wrapper = _execute(self._wrapper or self._getter)
        return self._wrapper, self.provided


class ContextThreadPoolExecutor(th.ThreadPoolExecutor):
    def submit(self, fn, *args, **kwargs):
        context = ctx.get_current()
        # Get the list of locks held by this thread, we don't want sub
        # tasks locking the same thing!
        held_locks = list(_get_locks())

        def wrapped():
            # This is executed in the new thread.
            if context is not None:
                context.update_store()
            # Ensure the sub task knows about the parent's locks and doesn't
            # block on them.
            _set_locks(held_locks)
            return fn(*args, **kwargs)
        return super(ContextThreadPoolExecutor, self).submit(wrapped)


class FeedTask(tf_task.Task):
    """Invokes WrapperTasks in parallel over each EntryWrapper in a feed.

    Usage
      Creation:
        # Preferred
        fm = FeedTask('lpar_frobnicate', LPAR.getter(adapter))
      or
        # Non-preferred.  See 'Greedy Methods' warning below
        feed = LPAR.wrap(adapter.read(LPAR.schema_type, ...))
        fm = FeedTask('lpar_frobnicate', feed)

      Adding Subtasks:
        # Preferred
        fm.add_subtask(FrobnicateLpar(foo, bar))
        fm.add_functor_subtask(frobnify, abc, xyz)
      and/or
        # Non-preferred.  See 'Greedy Methods' warning below
        for uuid, txn in fm.wrapper_tasks.items():
            if meets_criteria(txn.wrapper, uuid):
                txn.add_subtask(FrobnicateLpar(baz, blah))
        fm.wrapper_tasks[known_uuid].add_subtask(FrobnicateLpar(baz, blah)

      Execution/TaskFlow management:
        main_flow.add(fm)
        ...
        taskflow.engines.run(main_flow)

    Warning: Greedy Methods
    This implementation makes every effort to defer the feed GET as long as
    possible.  The more time passes between the GET and the execution of the
    WrapperTasks, the more likely it is that some out-of-band change will have
    modified one of the objects represented in the feed. This will cause an
    etag mismatch on that WrapperTask's update (POST), resulting in that
    WrapperTask being redriven, which costs an extra GET+POST to the REST
    server.

    Consumers of this class can thwart these efforts by:
    a) Initializing the FeedTask with an already-retrieved feed instead of a
       FeedGetter; or
    b) Using any of the following methods/properties prior to execution.  All
       of these will trigger a GET of the feed if not already fetched:

    .wrapper_tasks
    .get_wrapper(uuid)
    .feed

    The cost is incurred only the first time one of these is used.  If your
    workflow requires calling one of these early, it is not necessary to
    avoid them subsequently.
    """
    def __init__(self, name, feed_or_getter, max_workers=10,
                 update_timeout=-1):
        """Create a FeedTask with a FeedGetter (preferred) or existing feed.

        :param name: A descriptive string name.  This will be used along with
                     each wrapper's UUID to generate the name for that
                     wrapper's WrapperTask.
        :param feed_or_getter: pypowervm.wrappers.entry_wrapper.FeedGetter or
                               an already-fetched feed (list of EntryWrappers)
                               over which to operate.
        :param max_workers: (Optional) Integer indicating the maximum number of
                            worker threads to run in parallel within the .flow
                            or by the .execute method. See
                            concurrent.futures.ThreadPoolExecutor(max_workers).
        :param update_timeout: (Optional) Integer number of seconds after which
                               to time each WrapperTask's POST request.  -1,
                               the default, causes the request to use the
                               timeout value configured on the Session
                               belonging to the Adapter.
        """
        super(FeedTask, self).__init__(name)
        if isinstance(feed_or_getter, ewrap.FeedGetter):
            self._feed = None
            self._getter = feed_or_getter
        elif isinstance(feed_or_getter, list):
            # Make sure the feed has something in it.
            if len(feed_or_getter) == 0:
                raise ex.FeedTaskEmptyFeed()
            # Make sure it's a list of EntryWrapper
            if [i for i in feed_or_getter
                    if not isinstance(i, ewrap.EntryWrapper)]:
                raise ValueError("List must contain EntryWrappers "
                                 "exclusively.")
            self._feed = feed_or_getter
            self._getter = None
        else:
            raise ValueError(_("Must supply either a list of EntryWrappers or "
                               "a FeedGetter."))
        # Max WrapperTasks to run in parallel
        self.max_workers = max_workers
        self.update_timeout = update_timeout
        # Map of {uuid: WrapperTask}.  We keep this empty until we need the
        # individual WraperTasks.  This is triggered by .wrapper_tasks and
        # .get_wrapper(uuid) (and obviously executing).
        self._tx_by_uuid = {}
        # Until we *need* individual WrapperTasks, save subtasks in one place.
        # EntryWrapperGetter is a cheat to allow us to build the WrapperTask.
        self._common_tx = WrapperTask(
            'internal', ewrap.EntryWrapperGetter(None, ewrap.Wrapper,
                                                 None))
        self._post_exec = []

    @property
    def wrapper_tasks(self):
        """(Greedy) Dictionary of {uuid: WrapperTask} for all wrappers.

        The first access of this property triggers a GET of the feed if it has
        not already been fetched, so use judiciously.
        """
        if not self._tx_by_uuid:
            # Create a separate WrapperTask for each wrapper in the feed.
            # As long as the consumer uses FeedTask.add_[functor_]subtask
            # and doesn't ask for .wrapper_tasks, we keep only one copy of the
            # subtask list.  Once the consumer "breaks the seal" and requests
            # individual WrapperTasks per wrapper, we need to (GET the feed -
            # this is triggered by .feed - and) create them based on this
            # common subtask list.
            # This is only done once.  Thereafter, .add_[functor_]subtask will
            # add separately to each WrapperTask.
            for entry in self.feed:
                name = '%s_%s' % (self.name, entry.uuid)
                self._tx_by_uuid[entry.uuid] = WrapperTask(
                    name, entry, subtasks=self._common_tx.subtasks,
                    allow_empty=True, update_timeout=self.update_timeout)
        return self._tx_by_uuid

    def get_wrapper(self, uuid):
        """(Greedy) Returns the EntryWrapper associated with a particular UUID.

        Note that this method triggers a GET of the feed if it has not already
        been fetched, so use judiciously.

        :param uuid: The UUID of the wrapper of interest.
        :return: The EntryWrapper instance with the specified UUID.
        :raise KeyError: If there's no WrapperTask for a wrapper with the
                         specified UUID.
        """
        # Grab it from the WrapperTask map (O(1)) rather than the feed (O(n)).
        # It'll also be up to date without having to trigger a feed rebuild.
        return self.wrapper_tasks[uuid].wrapper

    def add_subtask(self, task):
        """Add a Subtask to *all* WrapperTasks in this FeedTask.

        To add Subtasks to individual WrapperTasks, iterate over the result of
        the 'wrapper_tasks' property.

        Specification is the same as for WrapperTask.add_subtask.
        """
        if self._tx_by_uuid:
            # _tx_by_uuid is guaranteed to have WrapperTasks for all UUIDs,
            # including this one
            for txn in self._tx_by_uuid.values():
                txn.add_subtask(task)
        else:
            self._common_tx.add_subtask(task)
        return self

    def add_functor_subtask(self, func, *args, **kwargs):
        """Add a functor Subtask to *all* WrapperTasks in this FeedTask.

        To add Subtasks to individual WrapperTasks, iterate over the result of
        the 'wrapper_tasks' property.

        Specification is the same as for WrapperTask.add_functor_subtask.
        """
        return self.add_subtask(_FunctorSubtask(func, *args, **kwargs))

    def add_post_execute(self, *tasks):
        """Add some number of TaskFlow Tasks to run after the WrapperTasks.

        Such Tasks may 'require' a parameter called wrapper_task_rets, which
        will be a dict of the form:
        {uuid: {
            'wrapper': wrapper,
            label1: return_value,
            label2: return_value,
            ...
            labelN: return_value}}
        ...where:
        uuid is the UUID of the WrapperTask's wrapper.
        wrapper is the WrapperTask's wrapper in its final (possibly-updated)
                form.
        labelN: return_value are the return values from Subtasks using the
                'provides' mechanism.  Each label corresponds to the name
                given by the Subtask's 'provides' argument.

        :param tasks: Some number of TaskFlow Tasks (or Flows) to be executed
                      linearly after the parallel WrapperTasks have completed.
        """
        self._post_exec.extend(tasks)

    @property
    def feed(self):
        """(Greedy) Returns this FeedTask's feed (list of wrappers).

        The first access of this property triggers a GET of the feed if it has
        not already been fetched, so use this only if you need the
        EntryWrappers outside of the execution itself.
        """
        if self._feed is None:
            self._feed = self._getter.get()
        if len(self._feed) == 0:
            raise ex.FeedTaskEmptyFeed()
        # Do we need to refresh the feed based on having been run?
        # If we haven't replicated WrapperTasks yet, there's no chance we're
        # out of sync - and we don't want to trigger GET/replication.
        if self._tx_by_uuid:
            # Rebuild the entire feed from the WrapperTasks' .wrappers.
            # TAG_WRAPPER_SYNC
            # Note that, if this happens while the WrapperTasks are running,
            # we may be grabbing the wrapper from a WrapperTask "while" it is
            # being changed as the result of an update(). This is threadsafe as
            # long as the assignment (by WrapperTask.execute) and the accessor
            # (WrapperTask.wrapper) remain atomic by using simple =/return.
            for wrap in self._feed:
                if self.get_wrapper(wrap.uuid).etag != wrap.etag:
                    # Refresh needed
                    self._feed = [tx.wrapper for tx in
                                  self.wrapper_tasks.values()]
                    break
        return self._feed

    @staticmethod
    def _process_subtask_rets(subtask_rets):
        """Reshape the dict of wrapper_{uuid} and subtask_rets_{uuid}.

        Input form:  {'wrapper_%(uuid)s': EntryWrapper,
                      'subtask_rets_%(uuid)s': {
                            label1: return_value,
                            label2: return_value,
                            ...,
                            labelN: return_value}}

        Output form: {uuid: {
                            'wrapper': EntryWrapper,
                            label1: return_value,
                            label2: return_value,
                            ...,
                            labelN: return_value}}
        """
        ret = {}
        for key, val in subtask_rets.items():
            label, uuid = key.rsplit('_', 1)
            if label != 'wrapper':
                ret[uuid] = dict(val,
                                 wrapper=subtask_rets['wrapper_%s' % uuid])
        return ret

    def execute(self):
        """Run this FeedTask's WrapperTasks in parallel TaskFlow engine.

        :return: Dictionary of results provided by subtasks and post-execs.
                 The shape of this dict is as normally expected from TaskFlow,
                 noting that the WrapperTasks are executed in a subflow and
                 their results processed into wrapper_task_rets.  For example:
            {'wrapper_task_rets': { uuid: {...}, uuid: {...}, ...}
             'post_exec_x_provides': ...,
             'post_exec_y_provides': ...,
             ...}
        """
        # Ensure a true no-op (in particular, we don't want to GET the feed) if
        # there are no Subtasks
        if not any([self._tx_by_uuid, self._common_tx.subtasks,
                    self._post_exec]):
            LOG.info(_("FeedTask %s has no Subtasks; no-op execution."),
                     self.name)
            return
        rets = {'wrapper_task_rets': {}}
        try:
            # Calling .wrapper_tasks will cause the feed to be fetched and
            # WrapperTasks to be replicated, if not already done.  Only do this
            # if there exists at least one WrapperTask with Subtasks.
            # (NB: It is legal to have a FeedTask that *only* has post-execs.)
            if self._tx_by_uuid or self._common_tx.subtasks:
                pflow = tf_uf.Flow("%s_parallel_flow" % self.name)
                pflow.add(*self.wrapper_tasks.values())
                # Execute the parallel flow now so the results can be provided
                # to any post-execs.
                rets['wrapper_task_rets'] = self._process_subtask_rets(
                    tf_eng.run(
                        pflow, engine='parallel',
                        executor=ContextThreadPoolExecutor(self.max_workers)))
            if self._post_exec:
                flow = tf_lf.Flow('%s_post_execs' % self.name)
                flow.add(*self._post_exec)
                eng = tf_eng.load(flow, store=rets)
                eng.run()
                rets = eng.storage.fetch_all()
        except tf_ex.WrappedFailure as wfail:
            LOG.error(_("FeedTask %s experienced multiple exceptions. They "
                        "are logged individually below."), self.name)
            for fail in wfail:
                LOG.exception(fail.pformat(fail.traceback_str))
            raise ex.MultipleExceptionsInFeedTask(self.name, wfail)

        # Let a non-wrapped exception (which happens if there's only one
        # element in the feed) bubble up as-is.

        return rets



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pypowervm.utils import transaction as trans
from pypowervm.wrappers import management_console as mc


def get_public_key(adapter):
    """Get the public key for the management console.

    :param adapter: The adapter for the pypowervm API.
    :return: The public key
    """
    # Get the consoles feed and use the first, there is just one.
    console = mc.ManagementConsole.wrap(
        adapter.read(mc.ManagementConsole.schema_type))[0]
    return console.ssh_public_key


def add_authorized_key(adapter, public_key):
    """Add an authorized public key to the management console.

    The public_key will be added if it doesn't already exist.

    :param adapter: The adapter for the pypowervm API.
    :param public_key: The public key to be added.
    """
    console_w = mc.ManagementConsole.wrap(
        adapter.read(mc.ManagementConsole.schema_type))[0]

    @trans.entry_transaction
    def run_update(console):
        keys = console.ssh_authorized_keys
        if public_key not in keys:
            keys = list(keys)
            keys.append(public_key)
            console.ssh_authorized_keys = keys
            console.update()

    run_update(console_w)


def get_authorized_keys(adapter):
    """Get all authorized keys on the management console.

    :param adapter: The adapter for the pypowervm API.
    """
    console_w = mc.ManagementConsole.wrap(
        adapter.read(mc.ManagementConsole.schema_type))[0]

    return console_w.ssh_authorized_keys



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pypowervm import adapter as adpt
from pypowervm import const as c
from pypowervm import exceptions as pvm_exc
from pypowervm.tests.test_utils import pvmhttp
import pypowervm.wrappers.job as job


def load_file(file_name, adapter=None):
    """Helper method to load the responses from a given location."""
    return pvmhttp.load_pvm_resp(file_name, adapter).get_response()


def raiseRetryException():
    """Used for other tests wishing to raise an exception to a force retry."""
    resp = adpt.Response('reqmethod', 'reqpath', c.HTTPStatus.ETAG_MISMATCH,
                         'reason', 'headers')
    http_exc = pvm_exc.HttpError(resp)
    raise http_exc


def get_parm_checker(test_obj, exp_uuid, exp_job_parms, exp_job_mappings=[],
                     exp_timeout=None):
    # Utility method to return a dynamic parameter checker for tests

    # Build the expected job parameter strings
    exp_job_parms_str = [job.Job.create_job_parameter(k, v).toxmlstring()
                         for k, v in exp_job_parms]
    exp_job_parms_str += [
        job.Job.create_job_parameter(k, ",".join(v)).toxmlstring()
        for k, v in exp_job_mappings]

    def parm_checker(uuid, job_parms=None, timeout=None):
        # Check simple parms
        test_obj.assertEqual(exp_uuid, uuid)
        test_obj.assertEqual(exp_timeout, timeout)

        # Check the expected and actual number of job parms are equal
        test_obj.assertEqual(len(exp_job_parms_str), len(job_parms))

        # Ensure each parameter is in the list of expected.
        for parm in job_parms:
            test_obj.assertIn(parm.toxmlstring(), exp_job_parms_str)

    # We return our custom checker
    return parm_checker



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utility decorator to retry the decorated method."""

import functools
from oslo_log import log as logging
import random
from six import moves
import time

from pypowervm import const
from pypowervm import exceptions as exc
from pypowervm.i18n import _


LOG = logging.getLogger(__name__)

DFT_RETRY_CODES = frozenset([const.HTTPStatus.ETAG_MISMATCH])

NO_TEST = lambda *args, **kwds: True
NO_CHECKER = lambda *args, **kwds: False
NO_DELAY = lambda *args, **kwds: None
NO_ARGMOD = lambda this_try, max_tries, *args, **kwds: (args, kwds)

# Used by STEPPED_RANDOM_DELAY.  Each entry corresponds to the kwargs for
# gen_random_delay.  There's no magic to these numbers; they're fairly
# arbitrary.
RANDOM_DELAY_STEPS = ({'max_s': 0},
                      {'max_s': 1},
                      {'min_s': 0.5, 'max_s': 4},
                      {'min_s': 2, 'max_s': 13},
                      {'min_s': 6.5, 'max_s': 30},
                      # Subsequent steps 0-60s
                      {'max_s': 60})


def STEPPED_DELAY(attempt, max_attempts, *args, **kwds):
    """A delay function that increases its delay per attempt.

    The steps will be:
     - Attempt 1: 0.0s
     - Attempt 2: 0.5s
     - Attempt 3: 2.0s
     - Attempt 4: 6.5s
     - Attempt 5: 20s
     - Attempt 6+: 30s
    """
    sleep_time = (0.25 * (3**(attempt-1)) - 0.25)
    time.sleep(min(sleep_time, 30))


def gen_random_delay(min_s=0, max_s=10):
    """Generate a delay function that waits a random amount of time.

    :param min_s: Minimum number of seconds to delay (float).
    :param max_s: Maximum number of seconds to delay (float).
    :return: A delay method suitable for passing to retry's delay_func kwarg.
    """
    def RANDOM_DELAY(attempt, max_attempts, *args, **kwargs):
        span = max_s - min_s
        sleep_time = min_s + (random.random() * span)
        time.sleep(sleep_time)
    return RANDOM_DELAY


def STEPPED_RANDOM_DELAY(attempt, max_attempts, *args, **kwargs):
    """A delay function for increasing random sleep times.

    The RANDOM_DELAY_STEPS variable is used to determine the min/max for each
    step.  This is a graduating scale - based on the overall max_attempts
    specified.  If there are 60 max attempts, the first 10 will use the
    first delay in RANDOM_DELAY_STEPS, the next ten will use the second delay
    in RANDOM_DELAY_STEPS, etc...

    If there are only 6 retries, then the first will use the first position
    in RANDOM_DELAY_STEPS, the second will map to the second RANDOM_DELAY_STEPS
    and so on.
    """
    # Generate the position, based off the max attempts and the current pos
    pos = int(((attempt - 1) * len(RANDOM_DELAY_STEPS)) / max_attempts)

    # If for some reason, the user goes above the max attempts, limit it to
    # to the last position.
    pos = (pos if pos < len(RANDOM_DELAY_STEPS)
           else (len(RANDOM_DELAY_STEPS) - 1))

    # Run the random delay function
    gen_random_delay(**RANDOM_DELAY_STEPS[pos])(attempt, max_attempts, *args,
                                                **kwargs)


def refresh_wrapper(trynum, maxtries, *args, **kwargs):
    """A @retry argmod_func to refresh a Wrapper, which must be the first arg.

    When using @retry to decorate a method which modifies a Wrapper, a common
    cause of retry is etag mismatch.  In this case, the retry should refresh
    the wrapper before attempting the modifications again.  This method may be
    passed to @retry's argmod_func argument to effect such a refresh.

    Note that the decorated method must be defined such that the wrapper is its
    first argument.
    """
    arglist = list(args)
    # If we get here, we *usually* have an etag mismatch, so specifying
    # use_etag=False *should* be redundant.  However, for scenarios where we're
    # retrying for some other reason, we want to guarantee a fresh fetch to
    # obliterate any local changes we made to the wrapper (because the retry
    # should be making those changes again).
    arglist[0] = arglist[0].refresh(use_etag=False)
    return arglist, kwargs


def retry(tries=3, delay_func=NO_DELAY,
          retry_except=None, http_codes=DFT_RETRY_CODES, test_func=None,
          resp_checker=NO_CHECKER, limit_except=None, argmod_func=NO_ARGMOD):
    """Retry method decorator.

    :param tries: The max number of calls to the wrapped method.
    :param delay_func: A method to delay before retrying.
        Defaults to no delay.
        The parameters that are sent are:
            - the number of the current try
            - the maximum number of tries
            - the arguments to the decorated method
            - the keyword arguments to the decorated method
        No return value is expected.
    :param retry_except: An exception class (or tuple thereof) to retry if
        received.  Defaults to no exceptions besides the HttpError which is
        handled separately by the http_codes parameter.
    :param http_codes: A list of http response codes to retry if received.
        Default is to not handle any specific http codes.
    :param test_func: A method to call to determine whether to retry. This
        method takes precedence over http codes. That is, if specified, the
        http codes are not considered.
        The parameters that are sent are:
            - the exception that was received
            - the number of the current try
            - the maximum number of tries
            - the arguments to the decorated method
            - the keyword arguments to the decorated method
        The return value is expected to be boolean, True or False, where
            True means to retry the decorated method.
    :param resp_checker: A method to call when no exception is caught, to
        check the response and determine whether to retry.
        The parameters that are sent are:
            - the number of the current try
            - the maximum number of tries
            - the arguments to the decorated method
            - the keyword arguments to the decorated method
        The return value is expected to be boolean, True or False, where
            True means to retry the decorated method.
    :param limit_except: An exception to raise if the number of tries is
        exhausted.
    :param argmod_func: A method to call after delay_func, before retrying, to
                        modify the arguments to the main method.  The input
                        parameters are:
                            - the number of the current try
                            - the maximum number of tries
                            - the non-keyword arguments to the decorated method
                            - the keyword arguments to the decorated method
                        The return is expected to a list and a dict of the
                        new arguments to the decorated method.
                        Example:
                        def argmod(t, m, *a, **k):
                            l = list(a)
                            l[0] += 1
                            k['foo'] = bar
                            return l, k
    :returns: The return value of the wrapped method.
    """
    def _retry(func):
        @functools.wraps(func)
        def __retry(*args, **kwds):
            def _raise_exc():
                if _limit_except:
                    raise _limit_except
                else:
                    raise

            def _test_retry(e):
                # Determine if an exception should be raised
                if (not _test_func(e, try_, _tries, *args, **kwds) or
                        try_ == _tries):
                    _raise_exc()
                # Otherwise, we will continue trying
                return

            def _log_response_retry(try_, max_tries, uri, resp_code):
                LOG.warning(_('Attempt %(retry)d of total %(total)d for URI '
                              '%(uri)s.  Error was a known retry response '
                              'code: %(resp_code)s'),
                            {'retry': try_, 'total': max_tries, 'uri': uri,
                             'resp_code': resp_code})

            def _log_exception_retry(try_, max_tries, exc):
                LOG.warning(_('Attempt %(retry)d of %(total)d failed.  Will '
                              'retry. The exception was:\n %(except)s.'),
                            {'retry': try_, 'total': max_tries, 'except': exc})

            # Standardize input
            # For some reason, if we use the parms in an 'if' directly
            # python throws an exception.  Assigning them avoids it.
            _tries = tries
            _retry_except = retry_except
            _http_codes = http_codes
            _test_func = test_func
            _resp_checker = resp_checker
            _limit_except = limit_except
            _argmod_func = argmod_func

            if _retry_except is None:
                _retry_except = ()
            if _http_codes is None:
                _http_codes = ()
            caller_test_func = _test_func is not None
            if not caller_test_func:
                _test_func = NO_TEST
            if _resp_checker is None:
                _resp_checker = NO_CHECKER
            # Start retries
            for try_ in moves.range(1, _tries+1):
                try:
                    resp = func(*args, **kwds)
                    # No exception raised, call the response checker
                    # If we're on the last iteration, we return the response.
                    # The response checker should raise an exception if
                    # it doesn't want this behavior.
                    if (not _resp_checker(resp, try_, _tries, *args, **kwds)
                            or try_ == _tries):
                        return resp
                except exc.HttpError as e:
                    if caller_test_func or e.response.status in _http_codes:
                        _test_retry(e)
                        _log_response_retry(try_, _tries, e.response.reqpath,
                                            e.response.status)
                    else:
                        _raise_exc()
                except _retry_except as e:
                    _test_retry(e)
                    _log_exception_retry(try_, _tries, e)
                # If we get here then we're going to retry
                delay_func(try_, _tries, *args, **kwds)
                # Adjust arguments if necessary
                args, kwds = _argmod_func(try_, _tries, *args, **kwds)
        return __retry
    return _retry



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from pypowervm.utils import validation as vldn
from pypowervm.wrappers import base_partition as bp
from pypowervm.wrappers import logical_partition as lpar
from pypowervm.wrappers import managed_system as mgd_sys


class TestValidator(testtools.TestCase):
    """Unit tests for validation."""

    def setUp(self):
        super(TestValidator, self).setUp()

        def _bld_mgd_sys(proc_units_avail=20.0, mem_free=32768,
                         system_name='default_sys_name',
                         max_procs_per_aix_linux_lpar=10,
                         max_sys_procs_limit=15,
                         max_vcpus_per_aix_linux_lpar=10,
                         max_sys_vcpus_limit=15):
            # Build a fake managed system wrapper
            mngd_sys = mock.MagicMock()
            mngd_sys.system_name = system_name
            mngd_sys.proc_units_avail = proc_units_avail
            mngd_sys.memory_free = mem_free
            mngd_sys.max_procs_per_aix_linux_lpar = (
                max_procs_per_aix_linux_lpar)
            mngd_sys.max_sys_procs_limit = (
                max_sys_procs_limit)
            mngd_sys.max_vcpus_per_aix_linux_lpar = (
                max_vcpus_per_aix_linux_lpar)
            mngd_sys.max_sys_vcpus_limit = (
                max_sys_vcpus_limit)
            mocked = mock.MagicMock(spec=mgd_sys.System, return_value=mngd_sys)
            return mocked()

        def _bld_lpar(proc_units=1.0, min_mem=512, des_mem=2048, max_mem=4096,
                      has_dedicated=False, name='default', rmc_state='active',
                      mem_dlpar=True, proc_dlpar=True, state='running',
                      env='AIX/Linux', proc_compat='Default', srr_enabled=True,
                      min_vcpus=1, des_vcpus=2, max_vcpus=4,
                      min_proc_units=0.1, max_proc_units=1.0, pool_id=None,
                      exp_factor=0.0, ame_enabled=False):
            lpar_w = mock.MagicMock()
            # name, states, env, etc.
            lpar_w.name = name
            lpar_w.state = state
            lpar_w.rmc_state = rmc_state
            lpar_w.env = env
            lpar_w.proc_compat_mode = proc_compat
            lpar_w.srr_enabled = srr_enabled
            # Proc
            lpar_w.proc_config.has_dedicated = has_dedicated
            if has_dedicated:
                lpar_w.proc_config.dedicated_proc_cfg.desired = proc_units
                lpar_w.proc_config.dedicated_proc_cfg.max = max_vcpus
                lpar_w.proc_config.dedicated_proc_cfg.min = min_vcpus
            else:
                lpar_w.proc_config.shared_proc_cfg.desired_units = proc_units
                lpar_w.proc_config.shared_proc_cfg.desired_virtual = des_vcpus
                lpar_w.proc_config.shared_proc_cfg.max_virtual = max_vcpus
                lpar_w.proc_config.shared_proc_cfg.min_virtual = min_vcpus
                lpar_w.proc_config.shared_proc_cfg.pool_id = (
                    pool_id if pool_id else 0)
                lpar_w.proc_config.shared_proc_cfg.min_units = min_proc_units
                lpar_w.proc_config.shared_proc_cfg.max_units = max_proc_units
            # Mem
            lpar_w.mem_config.desired = des_mem
            lpar_w.mem_config.min = min_mem
            lpar_w.mem_config.max = max_mem
            lpar_w.mem_config.exp_factor = exp_factor
            # Can Modify
            if (state != bp.LPARState.NOT_ACTIVATED
               and rmc_state != bp.RMCState.ACTIVE):
                    lpar_w.can_modify_proc.return_value = (False, 'Bad RMC')
                    lpar_w.can_modify_mem.return_value = (False, 'Bad RMC')
            else:
                # Doesn't matter what the message is unless it's bad
                # so always make it bad
                lpar_w.can_modify_proc.return_value = (proc_dlpar,
                                                       'Bad proc DLPAR')
                lpar_w.can_modify_mem.return_value = (mem_dlpar,
                                                      'Bad mem DLPAR')
            mocked = mock.MagicMock(spec_set=lpar.LPAR, return_value=lpar_w)
            return mocked()

        self.mngd_sys = _bld_mgd_sys()
        self.lpar_21_procs = _bld_lpar(proc_units=21.0, name='lpar_21_procs')
        self.lpar_1_proc = _bld_lpar()
        self.lpar_11_vcpus = _bld_lpar(des_vcpus=11, name='11_vcpus')
        self.lpar_16_max_vcpus = _bld_lpar(max_vcpus=16, name='16_max_vcpus')
        self.lpar_1_proc_ded = _bld_lpar(has_dedicated=True, name='1_proc_ded')
        self.lpar_11_proc_ded = _bld_lpar(proc_units=11, has_dedicated=True,
                                          name='11_proc_ded')
        self.lpar_16_proc_max_ded = _bld_lpar(max_vcpus=16, has_dedicated=True,
                                              name='16_proc_max_ded')
        self.lpar_21_proc_ded = _bld_lpar(proc_units=21, has_dedicated=True,
                                          name='21_proc_ded')
        self.lpar_no_rmc = _bld_lpar(rmc_state='inactive')
        self.lpar_bad_mem_dlpar = _bld_lpar(mem_dlpar=False)
        self.lpar_bad_proc_dlpar = _bld_lpar(proc_dlpar=False)
        self.lpar_48g_mem = _bld_lpar(des_mem=48000, name='lpar_48g_mem')

        self.lpar_1_min_vcpus = _bld_lpar(min_vcpus=1, name='1_min_vcpus')
        self.lpar_2_min_vcpus = _bld_lpar(min_vcpus=2, name='2_min_vcpus')
        self.lpar_1_min_proc_units = _bld_lpar(min_proc_units=0.1,
                                               name='0.1_min_procs')
        self.lpar_3_min_proc_units = _bld_lpar(min_proc_units=0.3,
                                               name='0.3_min_procs')
        self.lpar_6_max_proc_units = _bld_lpar(max_proc_units=0.6,
                                               name='0.6_max_procs')
        self.lpar_9_max_proc_units = _bld_lpar(max_proc_units=0.9,
                                               name='0.9_max_procs')
        self.lpar_6_max_vcpus = _bld_lpar(max_vcpus=6, name='6_max_vcpus')
        self.lpar_8_max_vcpus = _bld_lpar(max_vcpus=8, name='8_max_vcpus')
        self.lpar_512mb_min_mem = _bld_lpar(min_mem=512, name='512_min_mem')
        self.lpar_1gb_min_mem = _bld_lpar(min_mem=1024, name='1gb_min_mem')
        self.lpar_6g_max_mem = _bld_lpar(max_mem=6144, name='6gb_max_mem')
        self.lpar_8g_max_mem = _bld_lpar(max_mem=8192, name='8gb_max_mem')
        self.lpar_default_spp = _bld_lpar(pool_id=0, name='default_spp')
        self.lpar_non_default_spp = _bld_lpar(pool_id=2,
                                              name='non_default_spp')
        self.lpar_power8_proc_compat = _bld_lpar(proc_compat="POWER8",
                                                 name='power8_compat_mode')
        self.lpar_srr_disabled = _bld_lpar(srr_enabled=False,
                                           name='srr_disabled')
        self.lpar_1_proc_ded_inactive = _bld_lpar(has_dedicated=True,
                                                  name='1_proc_ded_inactive',
                                                  state='not activated')
        self.lpar_22_procs = _bld_lpar(proc_units=22.0, name='lpar_22_procs')
        self.lpar_4_proc_ded = _bld_lpar(proc_units=4.0,
                                         has_dedicated=True, name='4_proc_ded')
        self.lpar_22_proc_ded = _bld_lpar(proc_units=22, has_dedicated=True,
                                          name='21_proc_ded')
        self.lpar_4g_mem = _bld_lpar(des_mem=4096, name='4gb_mem')
        self.lpar_6g_mem = _bld_lpar(des_mem=6144, name='6gb_mem')
        self.lpar_1dot6_proc_units = _bld_lpar(proc_units=1.6,
                                               name='1.6_procs')
        self.lpar_2dot2_proc_units = _bld_lpar(proc_units=2.2,
                                               name='2.2_procs')
        self.lpar_1_vcpus = _bld_lpar(des_vcpus=1, name='lpar_1_vcpus')
        self.lpar_not_activated = _bld_lpar(name='lpar_not_activated',
                                            state='not activated')
        self.lpar_running = _bld_lpar(name='lpar_running', state='running')
        self.lpar_starting = _bld_lpar(name='lpar_starting', state='starting')
        self.lpar_ame_2 = _bld_lpar(name='ame_2', exp_factor=2.0,
                                    ame_enabled=True)
        self.lpar_ame_3 = _bld_lpar(name='ame_3', exp_factor=3.0,
                                    ame_enabled=True)

    def test_validator(self):
        # Test desired proc units > host avail proc units fails for shared
        vldr = vldn.LPARWrapperValidator(self.lpar_21_procs, self.mngd_sys)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)

        # Test desired proc units < host avail proc units passes for shared
        vldn.LPARWrapperValidator(self.lpar_1_proc,
                                  self.mngd_sys).validate_all()

        # Test desired proc units > host avail proc units fails for dedicated
        vldr = vldn.LPARWrapperValidator(self.lpar_21_proc_ded, self.mngd_sys)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)

        # Test desired proc units < host avail proc units passes for dedicated
        vldn.LPARWrapperValidator(self.lpar_1_proc_ded,
                                  self.mngd_sys).validate_all()

        # Test resize fails with inactive rmc
        vldr = vldn.LPARWrapperValidator(self.lpar_1_proc, self.mngd_sys,
                                         cur_lpar_w=self.lpar_no_rmc)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test resize fails with no mem dlpar
        vldr = vldn.LPARWrapperValidator(self.lpar_1_proc, self.mngd_sys,
                                         cur_lpar_w=self.lpar_bad_mem_dlpar)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test resize fails with no proc dlpar
        vldr = vldn.LPARWrapperValidator(self.lpar_1_proc, self.mngd_sys,
                                         cur_lpar_w=self.lpar_bad_proc_dlpar)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)

        # Test dedicated procs > host max allowed procs per lpar fails
        vldr = vldn.LPARWrapperValidator(self.lpar_11_proc_ded, self.mngd_sys)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test dedicated max procs > host max sys procs limit fails
        vldr = vldn.LPARWrapperValidator(self.lpar_16_proc_max_ded,
                                         self.mngd_sys)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test shared desired vcpus > host max allowed vcpus per lpar fails
        vldr = vldn.LPARWrapperValidator(self.lpar_11_vcpus, self.mngd_sys)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test shared desired max vcpus > host max sys vcpus limit fails
        vldr = vldn.LPARWrapperValidator(self.lpar_16_max_vcpus, self.mngd_sys)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)

        # Test desired memory > host available memory fails
        vldr = vldn.LPARWrapperValidator(self.lpar_48g_mem, self.mngd_sys)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)

        # Test changing min vcpus fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_1_min_vcpus, self.mngd_sys,
                                         cur_lpar_w=self.lpar_2_min_vcpus)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing max vcpus fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_8_max_vcpus, self.mngd_sys,
                                         cur_lpar_w=self.lpar_6_max_vcpus)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing min proc units fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_3_min_proc_units,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_min_proc_units)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing max proc units fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_9_max_proc_units,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_6_max_proc_units)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing min memory fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_512mb_min_mem,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_1gb_min_mem)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing max memory fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_8g_max_mem, self.mngd_sys,
                                         cur_lpar_w=self.lpar_6g_max_mem)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing AME expansion factor from 2 to 3 fails active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_ame_3, self.mngd_sys,
                                         cur_lpar_w=self.lpar_ame_2)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test toggling AME fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_ame_2, self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_proc)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test resizing lpar from defaultSPP to non-defaultSPP passes
        vldr = vldn.LPARWrapperValidator(self.lpar_non_default_spp,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_default_spp)
        vldr.validate_all()
        # Test resizing lpar from non-defaultSPP to defaultSPP passes
        vldr = vldn.LPARWrapperValidator(self.lpar_default_spp,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_non_default_spp)
        vldr.validate_all()
        # Test changing from dedicated to non-defaultSPP passes
        vldr = vldn.LPARWrapperValidator(self.lpar_non_default_spp,
                                         self.mngd_sys,
                                         self.lpar_1_proc_ded_inactive)
        vldr.validate_all()
        # Test changing processor mode (shared -> ded) fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_1_proc_ded,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_proc)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing processor mode (ded to shared) fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_1_proc,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_proc_ded)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing processor compatibility mode fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_power8_proc_compat,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_proc)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test changing SRR capabilty fails for active resize
        vldr = vldn.LPARWrapperValidator(self.lpar_srr_disabled,
                                         self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_proc)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test desired delta proc units > host avail proc units fails
        # during resize (shared -> shared)
        vldr = vldn.LPARWrapperValidator(self.lpar_22_procs, self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_proc)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test desired delta proc units <= host avail proc units passes
        # during resize (shared -> shared)
        vldn.LPARWrapperValidator(self.lpar_21_procs,
                                  self.mngd_sys,
                                  cur_lpar_w=self.lpar_1_proc).validate_all()
        # Test desired delta proc units > host avail proc units fails
        # during resize (dedicated -> dedicated)
        vldr = vldn.LPARWrapperValidator(self.lpar_22_proc_ded, self.mngd_sys,
                                         cur_lpar_w=self.lpar_1_proc_ded)
        self.assertRaises(vldn.ValidatorException, vldr.validate_all)
        # Test desired delta proc units <= host avail proc units passes
        # during resize (dedicated -> dedicated)
        vldn.LPARWrapperValidator(self.lpar_4_proc_ded,
                                  self.mngd_sys,
                                  self.lpar_1_proc_ded).validate_all()
        # Test resize delta mem
        mem_vldr = vldn.MemValidator(self.lpar_6g_mem, self.mngd_sys,
                                     cur_lpar_w=self.lpar_4g_mem)
        mem_vldr._populate_new_values()
        mem_vldr._populate_resize_diffs()
        self.assertEqual(2048, mem_vldr.delta_des_mem,
                         'Incorrect resize delta memory calculation')
        # Test resize delta procs
        proc_vldr = vldn.ProcValidator(self.lpar_4_proc_ded, self.mngd_sys,
                                       cur_lpar_w=self.lpar_1_proc_ded)
        proc_vldr._populate_new_values()
        proc_vldr._populate_resize_diffs()
        self.assertEqual(3, proc_vldr.delta_des_vcpus,
                         'Incorrect resize delta proc calculation'
                         ' in dedicated mode')
        proc_vldr = vldn.ProcValidator(self.lpar_2dot2_proc_units,
                                       self.mngd_sys,
                                       cur_lpar_w=self.lpar_1dot6_proc_units)
        proc_vldr._populate_new_values()
        proc_vldr._populate_resize_diffs()
        self.assertEqual(0.60, proc_vldr.delta_des_vcpus,
                         'Incorrect resize delta proc calculation in'
                         ' shared mode')
        proc_vldr = vldn.ProcValidator(self.lpar_1dot6_proc_units,
                                       self.mngd_sys,
                                       cur_lpar_w=self.lpar_1_proc_ded)
        proc_vldr._populate_new_values()
        proc_vldr._populate_resize_diffs()
        self.assertEqual(0.60, proc_vldr.delta_des_vcpus,
                         'Incorrect delta proc calculation while resizing '
                         'from dedicated to shared mode')
        proc_vldr = vldn.ProcValidator(self.lpar_4_proc_ded, self.mngd_sys,
                                       cur_lpar_w=self.lpar_1dot6_proc_units)
        proc_vldr._populate_new_values()
        proc_vldr._populate_resize_diffs()
        self.assertEqual(2.40, proc_vldr.delta_des_vcpus,
                         'Incorrect delta proc calculation while resizing '
                         'from shared to dedicated mode')
        # Test resizing not activated state lpar makes inactive_resize_checks
        with mock.patch('pypowervm.utils.validation.ProcValidator.'
                        '_validate_inactive_resize') as inactive_resize_checks:
            proc_vldr = vldn.ProcValidator(self.lpar_not_activated,
                                           self.mngd_sys,
                                           cur_lpar_w=self.lpar_not_activated)
            proc_vldr.validate()
            self.assertTrue(inactive_resize_checks.called,
                            'Inactive resize validations not performed.')
        # Test resizing running state lpar makes active_resize_checks
        with mock.patch('pypowervm.utils.validation.ProcValidator.'
                        '_validate_active_resize') as active_resize_checks:
            proc_vldr = vldn.ProcValidator(self.lpar_running, self.mngd_sys,
                                           cur_lpar_w=self.lpar_running)
            proc_vldr.validate()
            self.assertTrue(active_resize_checks.called,
                            'Active resize validations not performed.')
        # Test resizing starting state lpar makes active_resize_checks
        with mock.patch('pypowervm.utils.validation.ProcValidator.'
                        '_validate_active_resize') as active_resize_checks:
            proc_vldr = vldn.ProcValidator(self.lpar_starting, self.mngd_sys,
                                           cur_lpar_w=self.lpar_starting)
            proc_vldr.validate()
            self.assertTrue(active_resize_checks.called,
                            'Active resize validations not performed.')
